
# ==============================================================================
# Modules
# ==============================================================================
# Built-ins
import os, sys, time, multiprocessing, itertools, copy, datetime, warnings
from collections import OrderedDict, defaultdict

# PyData
import pandas as pd
import numpy as np
import xarray as xr
import networkx as nx

# Biology
from Bio import SeqIO, Seq
import ete3
import skbio

# SciPy
# from scipy import stats
from scipy.stats import entropy
from scipy.spatial.distance import squareform, pdist
from scipy.spatial import distance
try:
    from fastcluster import linkage
except ImportError:
    from scipy.cluster.hierarchy import linkage
    print("Could not import `linkage` from `fastcluster` and using `scipy.cluster.hierarchy.linkage` instead", file=sys.stderr)

# Statistics and Miscellaneous Machine Learning
# from astropy.stats import biweight_midcorrelation, median_absolute_deviation
from sklearn.metrics.pairwise import pairwise_distances

# Soothsayer
from ..transmute.conversion import linkage_to_newick
# from ..r_wrappers.packages.WGCNA import bicor
from ..utils import  infer_tree_type, is_symmetrical, force_symmetry, assert_acceptable_arguments, format_header, check_packages, is_dict_like, dict_build, is_number, is_nonstring_iterable, format_memory, name_tree_nodes, add_objects_to_globals

from ..io import write_object
import datetime, copy, warnings
from scipy.spatial.distance import squareform
from fastcluster import linkage
from soothsayer.transmute import linkage_to_newick
from skbio.util._decorator import experimental, stable
from scipy.stats import entropy

__all__ = {"pairwise", "pairwise_tree_distance","pairwise_difference", "pairwise_logfc"}

# compositional
import compositional as coda
functions_from_compositional = {"pairwise_vlr", "pairwise_rho","pairwise_phi"}
add_objects_to_globals(coda, functions_from_compositional, globals(), add_version=True, __all__=__all__)

# hive_networkx
import hive_networkx as hx
functions_from_hive_networkx= {"Symmetric", "dense_to_condensed", "condensed_to_dense"}
add_objects_to_globals(hx, functions_from_hive_networkx, globals(), add_version=True, __all__=__all__)


# ensemble_networkx
import ensemble_networkx as enx
functions_from_ensemble_networkx= { "pairwise_biweight_midcorrelation"}
add_objects_to_globals(enx, functions_from_ensemble_networkx, globals(), add_version=True, __all__=__all__)


__all__ = sorted(__all__)



# =======================================================
# Pairwise calculations
# =======================================================
# def dense_to_condensed(X, name=None, assert_symmetry=True, tol=None):
#     if assert_symmetry:
#         assert is_symmetrical(X, tol=tol), "`X` is not symmetric with tol=`{}`".format(tol)
#     labels = X.index
#     index=pd.Index(list(map(frozenset,itertools.combinations(labels, 2))), name=name)
#     data = squareform(X, checks=False)
#     return pd.Series(data, index=index, name=name)

# def condensed_to_dense(y:pd.Series, fill_diagonal=np.nan, index=None):
#     # Need to optimize this
#     data = defaultdict(dict)
#     for edge, w in y.iteritems():
#         node_a, node_b = tuple(edge)
#         data[node_a][node_b] = data[node_b][node_a] = w
        
#     if is_dict_like(fill_diagonal):
#         for node in data:
#             data[node][node] = fill_diagonal[node]
#     else:
#         for node in data:
#             data[node][node] = fill_diagonal
            
#     df_dense = pd.DataFrame(data)
#     if index is not None:
#         df_dense = df_dense.loc[index,index]
#     return df_dense

# # Symmetrical dataframes represented as augment pd.Series
# @experimental(as_of="2020.06.23")
# class Symmetric(object):
#     """
#     An indexable symmetric matrix stored as the lower triangle for space

#     devel
#     =====
#     2020-June-23
#     * Replace self._dense_to_condensed to dense_to_condensed
#     * Dropped math operations
#     * Added input for Symmetric or pd.Series with a frozenset index

#     2018-August-16
#     * Added __add__, __sub__, etc.
#     * Removed conversion to dissimilarity for tree construction
#     * Added .iteritems method
    

#     Future:
#     * Use `weights` instead of `data`
    
#     Dropped:
#     Fix the diagonal arithmetic
#     """
#     def __init__(self, data, node_type=None, edge_type=None,  func_metric=None, name=None, association="infer", assert_symmetry=True, nans_ok=True, tol=None, acceptable_associations={"similarity", "dissimilarity", "statistical_test", "network", "infer", None}, **attrs):
        
#         self._acceptable_associations = acceptable_associations
#         # From Symmetric object
#         if isinstance(data, type(self)):
#             self.__dict__.update(data.__dict__)
#             if (self.func_metric is None) and (func_metric is not None):
#                 assert hasattr(func_metric, "__call__"), "`func_metric` isn't a function"
#                 if (self.edge_type is None) and (edge_type is None):
#                     edge_type = func_metric.__name__
#             if (self.node_type is None) and (node_type is not None):
#                 self.node_type = node_type
#             if (self.edge_type is None) and (edge_type is not None):
#                 self.edge_type = edge_type
#             if (association  not in {"infer", None}):
#                 assert_acceptable_arguments(association, self._acceptable_associations)
#                 self.association = association

#         else:
#             # From pd.DataFrame object
#             if isinstance(data, pd.DataFrame):
#                 self._from_dataframe(data=data, association=association, assert_symmetry=assert_symmetry, nans_ok=nans_ok, tol=tol)

#             # From pd.Series object
#             if isinstance(data, pd.Series):
#                 self._from_series(data=data, association=association)
                
#             # From pd.DataFrame or pd.Series
#             self.metadata = dict()
#             if not nans_ok:
#                 assert not np.any(data.isnull()), "Cannot move forward with missing values"
#             if func_metric is not None:
#                 assert hasattr(func_metric, "__call__"), "`func_metric` isn't a function"
#                 if edge_type is None:
#                     edge_type = func_metric.__name__
#             self.node_type = node_type
#             self.edge_type = edge_type
#             self.func_metric = func_metric
#             self.values = self.weights.values
#             self.number_of_nodes = self.nodes.size
#             self.number_of_edges = self.edges.size
            
#         # Universal
#         self.name = name
# #         self.graph = self.to_graph(into=graph)
#         self.memory = self.weights.memory_usage()
#         self.metadata.update(attrs)
#         self.__synthesized__ = datetime.datetime.utcnow()
                                      
 

#     # =======
#     # Utility
#     # =======
#     def _infer_association(self, X):
#         diagonal = np.diagonal(X)
#         diagonal_elements = set(diagonal)
#         assert len(diagonal_elements) == 1, "Cannot infer relationships from diagonal because multiple values"
#         assert diagonal_elements <= {0,1}, "Diagonal should be either 0.0 for dissimilarity or 1.0 for similarity"
#         return {0.0:"dissimilarity", 1.0:"similarity"}[list(diagonal_elements)[0]]

            
#     def _from_dataframe(self, data:pd.DataFrame, association, assert_symmetry, nans_ok, tol):
#         if assert_symmetry:
#             assert is_symmetrical(data, tol=tol), "`X` is not symmetric.  Consider dropping the `tol` to a value such as `1e-10` or using `(X+X.T)/2` to force symmetry"
#         assert_acceptable_arguments(association, self._acceptable_associations)
#         if association == "infer":
#             association = self._infer_association(data)
#         assert_acceptable_arguments(association, self._acceptable_associations)
#         self.association = association
#         self.nodes = pd.Index(sorted(data.index), name="Nodes")
#         self.diagonal = pd.Series(np.diagonal(data), index=data.index, name="Diagonal")[self.nodes]
#         self.weights = dense_to_condensed(data, name="Weights", assert_symmetry=assert_symmetry, tol=tol)
#         self.edges = pd.Index(self.weights.index, name="Edges")

                                      
#     def _from_series(self, data:pd.Series, association):
#         assert np.all(data.index.map(lambda edge: isinstance(edge, frozenset))), "If `data` is pd.Series then each key in the index must be a frozenset of size 2"
#         if association == "infer":
#             association = None
#         assert_acceptable_arguments(association, self._acceptable_associations)
#         self.association = association
#         self.weights = pd.Series(data, name="Weights")
#         self.edges = pd.Index(self.weights.index, name="Edges")
#         self.nodes = pd.Index(sorted(frozenset.union(*self.edges)), name="Nodes")

        
#     def set_diagonal(self, diagonal):
#         if diagonal is None:
#             self.diagonal = None
#         else:
#             if is_number(diagonal):
#                 diagonal = dict_build([(diagonal, self.nodes)])
#             assert is_dict_like(diagonal), "`diagonal` must be dict-like"
#             assert set(diagonal.keys()) >= set(self.nodes), "Not all `nodes` are in `diagonal`"
#             self.diagonal =  pd.Series(diagonal, name="Diagonal")[self.nodes]
            
#     # =======
#     # Built-in
#     # =======
#     def __repr__(self):
#         pad = 4
#         header = format_header("Symmetric(Name:{}, dtype: {})".format(self.name, self.weights.dtype),line_character="=")
#         n = len(header.split("\n")[0])
#         fields = [
#             header,
#             pad*" " + "* Number of nodes ({}): {}".format(self.node_type, self.number_of_nodes),
#             pad*" " + "* Number of edges ({}): {}".format(self.edge_type, self.number_of_edges),
#             pad*" " + "* Association: {}".format(self.association),
#             pad*" " + "* Memory: {}".format(format_memory(self.memory)),
#             *map(lambda line:pad*" " + line, format_header("| Weights", "-", n=n-pad).split("\n")),
#             *map(lambda line: pad*" " + line, repr(self.weights).split("\n")[1:-1]),
#             ]

#         return "\n".join(fields)
    
#     def __getitem__(self, key):
#         """
#         `key` can be a node or non-string iterable of edges
#         """

#         if is_nonstring_iterable(key):
#             assert len(key) >= 2, "`key` must have at least 2 identifiers. e.g. ('A','B')"
#             key = frozenset(key)
#             if len(key) == 1:
#                 return self.diagonal[list(key)[0]]
#             else:
#                 if len(key) > 2:
#                     key = list(map(frozenset, itertools.combinations(key, r=2)))
#                 return self.weights[key]
#         else:
#             if key in self.nodes:
#                 s = frozenset([key])
#                 mask = self.edges.map(lambda x: bool(s & x))
#                 return self.weights[mask]
#             else:
#                 raise KeyError("{} not in node list".format(key))
        
#     @experimental(as_of="2020.06.23")
#     def __call__(self, key, func=np.sum):
#         """
#         This can be used for connectivity in the context of networks but can be confusing with the versatiliy of __getitem__
#         """
#         if hasattr(key, "__call__"):
#             return self.weights.groupby(key).apply(func)
#         else:
#             return func(self[key])
        
#     def __len__(self):
#         return self.number_of_nodes
#     def __iter__(self):
#         for v in self.weights:
#             yield v
#     def items(self):
#         return self.weights.items()
#     def iteritems(self):
#         return self.weights.iteritems()
#     def keys(self):
#         return self.weights.keys()
    
#     def apply(self, func):
#         return func(self.weights)
#     def mean(self):
#         return self.weights.mean()
#     def median(self):
#         return self.weights.median()
#     def min(self):
#         return self.weights.min()
#     def max(self):
#         return self.weights.max()
#     def idxmin(self):
#         return self.weights.idxmin()
#     def idxmax(self):
#         return self.weights.idxmax()
#     def sum(self):
#         return self.weights.sum()
#     def sem(self):
#         return self.weights.sem()
#     def var(self):
#         return self.weights.var()
#     def std(self):
#         return self.weights.std()
#     def describe(self, **kwargs):
#         return self.weights.describe(**kwargs)
#     def map(self, func):
#         return self.weights.map(func)
#     def entropy(self, base=2):
#         assert np.all(self.weights > 0), "All weights must be greater than 0"
#         return entropy(self.weights, base=base)

# #     # Maths
# #     def __add__(self, x):
# #         symmetric_clone = copy.deepcopy(self)
# #         symmetric_clone.data += x
# #         self.diagonal += x
# #         return symmetric_clone
# #     def __radd__(self, x):
# #         self.diagonal += x
# #         return Symmetric.__add__(self,x)
    
# #     def __sub__(self, x):
# #         symmetric_clone = copy.deepcopy(self)
# #         symmetric_clone.data -= x
# #         return symmetric_clone
# #     def __rsub__(self, x):
# #         symmetric_clone = copy.deepcopy(self)
# #         symmetric_clone.data = x - symmetric_clone.data
# #         return symmetric_clone
# #     def __mul__(self, x):
# #         symmetric_clone = copy.deepcopy(self)
# #         symmetric_clone.data *= x
# #         return symmetric_clone
# #     def __rmul__(self, x):
# #         return Symmetric.__mul__(self,x)
# #     def __truediv__(self, x):
# #         symmetric_clone = copy.deepcopy(self)
# #         symmetric_clone.data /= x
# #         return symmetric_clone
# #     def __rtruediv__(self, x):
# #         symmetric_clone = copy.deepcopy(self)
# #         symmetric_clone.data = x / symmetric_clone.data
# #         return symmetric_clone
# #     def __floordiv__(self, x):
# #         symmetric_clone = copy.deepcopy(self)
# #         symmetric_clone.data //= x
# #         return symmetric_clone
# #     def __rfloordiv__(self, x):
# #         symmetric_clone = copy.deepcopy(self)
# #         symmetric_clone.data = x // symmetric_clone.data
# #         return symmetric_clone

#     # ==========
#     # Conversion
#     # ==========
#     def to_dense(self, index=None):
#         return condensed_to_dense(y=self.weights, fill_diagonal=self.diagonal, index=index)

#     def to_condensed(self):
#         return self.weights

#     @check_packages(["ete3", "skbio"])
#     def to_tree(self, method="average", into=None, node_prefix="y"):
#         assert self.association == "dissimilarity", "`association` must be 'dissimilarity' to construct tree"
#         if method in {"centroid", "median", "ward"}:
#             warnings.warn("Methods ‘centroid’, ‘median’, and ‘ward’ are correctly defined only if Euclidean pairwise metric is used.\nSciPy Documentation - https://docs.scipy.org/doc/scipy/reference/generated/scipy.cluster.hierarchy.linkage.html#scipy.cluster.hierarchy.linkage") 
#         if into is None:
#             into = ete3.Tree
#         if not hasattr(self,"Z"):
#             self.Z = linkage(self.weights.values, metric="precomputed", method=method)
#         if not hasattr(self,"newick"):
#             self.newick = linkage_to_newick(self.Z, self.nodes)
#         tree = into(newick=self.newick, name=self.name)
#         return name_tree_nodes(tree, node_prefix)

#     def to_networkx(self, into=None, **attrs):
#         if into is None:
#             into = nx.Graph
#         metadata = {"name":self.name, "node_type":self.node_type, "edge_type":self.edge_type, "func_metric":self.func_metric}
#         metadata.update(attrs)
#         graph = into(**metadata)
#         for (node_A, node_B), weight in self.weights.iteritems():
#             graph.add_edge(node_A, node_B, weight=weight)
#         return graph
    
#     def to_file(self, path, **kwargs):
#         write_object(obj=self, path=path, **kwargs)

#     def copy(self):
#         return copy.deepcopy(self)

# Pairwise interactions
def pairwise(X, metric="euclidean", axis=1, name=None, into=pd.DataFrame, association="infer", signed=True, n_jobs=-1, check_metrics=True, symmetric_kws=dict()):
    """
    Compute pairwise interactions

    01-August-2018
    """

    def get_data_object(df_dense, into, association, func_signed, metric, symmetric_kws, CORRELATION_METRICS, DISTANCE_METRICS, name, check_metrics):
        if hasattr(metric, "__call__"):
            metric_name = metric.__name__
        else:
            metric_name = metric

        if association == "similarity":
            if check_metrics:
                assert metric_name not in DISTANCE_METRICS, f"Cannot compute simlilarity for {metric.__name__}"

        if association == "dissimilarity":
            if metric_name in CORRELATION_METRICS:
                df_dense = 1 - func_signed(df_dense)
        if into in [Symmetric, pd.Series]:
            kernel = Symmetric(df_dense, association=association, name=name, **symmetric_kws)
            if into == Symmetric:
                return kernel
            if into == pd.Series:
                return kernel.data
        else:
            return df_dense


    fill_diagonal="auto"
    CORRELATION_METRICS = {"kendalltau", "pearson", "spearman", 'biweight_midcorrelation', 'bicor'}
    STATISTICAL_TEST_METRICS = {'ks_2samp', 'ttest_1samp', 'ttest_ind', 'ttest_ind_from_stats', 'ttest_rel', "wilcoxon","ranksums", "mannwhitneyu"}
    DISTANCE_METRICS = distance.__dict__["__all__"]

    # Assertions
    accepted_associations = ["similarity", "dissimilarity", "statistical_test", "infer"]
    assert association in accepted_associations, f"{association} is not in {accepted_association}"

    # Metric name
    if hasattr(metric,"__call__"):
        metric_name = metric.__name__
    else:
        metric_name = metric
    # Infer association
    if association == "infer":
        if metric_name in CORRELATION_METRICS:
            association = "similarity"
        if metric_name in STATISTICAL_TEST_METRICS:
            association = "statistical_test"
        if metric_name in DISTANCE_METRICS:
            association = "dissimilarity"
        assert association != "infer", f"Unable to infer association from metric = {metric}"
        print(f"Inferred association as `{association}`", file=sys.stderr)

    # Similarity transformation function
    functions = {
        True: lambda X: (X+1)/2,
        False: lambda X:np.abs(X),
        None: lambda X:X,
    }
    func_signed = functions[signed]

    # Axis
    if axis == 0: X_copy = X.copy()
    if axis == 1: X_copy = X.copy().T

    # Locals
    Ar_X = X_copy.values
    labels = X_copy.index

    # Placeholder
    df_dense = None

    # Pairwise interactions
    while df_dense is None:
        # Statistical test
        if metric_name in STATISTICAL_TEST_METRICS:
            metric = lambda u,v: getattr(stats, metric_name)(u,v)[1]
            df_dense = pd.DataFrame(squareform(pdist(Ar_X, metric=metric)), index=labels, columns=labels)
            fill_diagonal = 0.0
            if association != "statistical_test":
                print(f"Convert `association` from `{association}` to `statistical_test`", file=sys.stderr)
                association = "statistical_test"
            break
        # Euclidean distance
        if metric_name == "euclidean":
            df_dense = pd.DataFrame(squareform(pdist(Ar_X, metric="euclidean")), index=labels, columns=labels)
            fill_diagonal = 0.0
            break
        # Biweight midcorrelation
        if metric in {"bicor", "biweight_midcorrelation"}:
            try:
                # df_dense = bicor(X_copy.T, condensed=False, self_interactions=True)
                df_dense = pairwise_biweight_midcorrelation(X_copy.T)
                break
            except:
                pass

        # Compute correlation and send the rest down the pipeline
        if metric_name in set(CORRELATION_METRICS) - set(["bicor", "biweight_midcorrelation"]):
            metric = metric_name
        if type(metric) == str:
            # Correlation
            if metric in CORRELATION_METRICS:
                df_dense = X_copy.T.corr(method = metric)
                break
            # Distance
            if metric_name in DISTANCE_METRICS:
                metric = getattr(distance, metric)

        # Check metric
        assert hasattr(metric, "__call__"), f"metric `{metric}` is unrecognized and should be converted into a callable at this point of the function"

        # Pairwise computation
        df_dense = pd.DataFrame(pairwise_distances(X_copy, metric=metric, n_jobs=n_jobs), index=labels, columns=labels)

        # Diagonal
        if fill_diagonal is not None:
            if fill_diagonal == "auto":
                fill_diagonal = {"similarity":1.0, "dissimilarity":0.0}[association]
            fill_diagonal = np.repeat(fill_diagonal, X_copy.shape[0]).tolist()
            np.fill_diagonal(df_dense.values, fill_diagonal)

    # Check dense
    assert df_dense is not None, f"`df_dense` is None.  Check the following: metric={metric}"

    # Check NaN
    df_isnull = df_dense.isnull()
    if np.any(df_isnull):
        num_null = df_isnull.values.ravel().sum()
        warnings.warn(f"There are {num_null} null values")

    # Symmetric keywords
    _symmetric_kws={
        "edge_type":metric_name,
        "func_metric":metric if hasattr(metric, "__call__") else None,
        "metadata":{
            "input_shape":X.shape,
        },
    }
    _symmetric_kws.update(symmetric_kws)

    # Return object
    return get_data_object(df_dense, into, association, func_signed, metric, _symmetric_kws, CORRELATION_METRICS, DISTANCE_METRICS, name, check_metrics)

# Pairwise distance on ete trees
def pairwise_tree_distance(tree, topology_only=True):
    d_topology = defaultdict(dict)
    leaves = tree.get_leaf_names()
    for i, id_obsv_A in enumerate(leaves):
        for j in range(i+1, len(leaves)):
            id_obsv_B = leaves[j]
            if id_obsv_A != id_obsv_B:
                d_topology[id_obsv_A][id_obsv_B] =  d_topology[id_obsv_B][id_obsv_A] = tree.get_distance(target=id_obsv_A, target2=id_obsv_B, topology_only=topology_only)
    df_dism_tree = pd.DataFrame(d_topology)
    np.fill_diagonal(df_dism_tree.values, 0)
    return df_dism_tree

# Pairwise difference
def pairwise_difference(X:pd.DataFrame, idx_ctrl, idx_treatment, name_ctrl="control", name_treatment="treatment"):
    """
    Positive values means they are greater in idx_treatment than in idx_ctrl.
    Negative values means they are greater in idx_ctrl than in idx_treatment.
    """
    # Init
    idx_attr = X.columns

    idx_ctrl = sorted(idx_ctrl)
    idx_treatment = sorted(idx_treatment)

    # Groups
    A_ctrl = X.loc[idx_ctrl,:].values
    A_treatment = X.loc[idx_treatment,:].values

    # Pairwise profiles
    diff_profiles = np.vstack(A_treatment[:, np.newaxis] - A_ctrl)

    # Labels
    idx_pairwise_labels = itertools.product(idx_treatment, idx_ctrl)
    names = [name_treatment, name_ctrl]
    return pd.DataFrame(diff_profiles, index=pd.MultiIndex.from_tuples(idx_pairwise_labels, names=names), columns=idx_attr)

# Pairwise log fold-change
def pairwise_logfc(X:pd.DataFrame, idx_ctrl, idx_treatment, func_log=np.log2, name_ctrl="control", name_treatment="treatment"):
    """
    Positive values means they are greater in idx_treatment than in idx_ctrl.
    Negative values means they are greater in idx_ctrl than in idx_treatment.
    """
    # return pairwise_difference(func_log(X), idx_ctrl=idx_ctrl, idx_treatment=idx_treatment, name_ctrl=name_ctrl, name_treatment=name_treatment) # MUST TEST THIS MORE FIRST BUT IT SHOULD WORK
    # Init
    idx_attr = X.columns

    idx_ctrl = sorted(idx_ctrl)
    idx_treatment = sorted(idx_treatment)

    # Log Transform
    X_logscale = func_log(X)

    # Groups
    A_ctrl = X_logscale.loc[idx_ctrl,:].values
    A_treatment = X_logscale.loc[idx_treatment,:].values

    # Pairwise profiles
    logfc_profiles = np.vstack(A_treatment[:, np.newaxis] - A_ctrl)

    # Labels
    idx_pairwise_labels = itertools.product(idx_treatment, idx_ctrl)
    names = [name_treatment, name_ctrl]
    return pd.DataFrame(logfc_profiles, index=pd.MultiIndex.from_tuples(idx_pairwise_labels, names=names), columns=idx_attr)

# # Biweight midcorrelation
# def pairwise_biweight_midcorrelation(X, use_numba=False, verbose=False):
#     """
#     X: {np.array, pd.DataFrame}

#     Code adapted from the following sources:
#         * https://stackoverflow.com/questions/61090539/how-can-i-use-broadcasting-with-numpy-to-speed-up-this-correlation-calculation/61219867#61219867
#         * https://github.com/olgabot/pandas/blob/e8caf4c09e1a505eb3c88b475bc44d9389956585/pandas/core/nanops.py

#     Special thanks to the following people:
#         * @norok2 (https://stackoverflow.com/users/5218354/norok2) for optimization (vectorization and numba)
#         * @olgabot (https://github.com/olgabot) for NumPy implementation

#     Benchmarking:
#         * iris_features (4,4)
#             * numba: 159 ms ± 2.85 ms per loop (mean ± std. dev. of 7 runs, 10 loops each)
#             * numpy: 276 µs ± 3.45 µs per loop (mean ± std. dev. of 7 runs, 1000 loops each)
#         * iris_samples: (150,150)
#             * numba: 150 ms ± 7.57 ms per loop (mean ± std. dev. of 7 runs, 10 loops each)
#             * numpy: 686 µs ± 18.8 µs per loop (mean ± std. dev. of 7 runs, 1000 loops each)

#     Future:
#         * Handle missing values

#     """
#     # Data
#     result = None
#     labels = None
#     if isinstance(X, pd.DataFrame):
#         labels = X.columns
#         X = X.values

#     def _base_computation(A):
#         n, m = A.shape
#         A = A - np.median(A, axis=0, keepdims=True)
#         v = 1 - (A / (9 * np.median(np.abs(A), axis=0, keepdims=True))) ** 2
#         est = A * v ** 2 * (v > 0)
#         norms = np.sqrt(np.sum(est ** 2, axis=0))
#         return n, m, est, norms

#     # Check if numba is available
#     assert_acceptable_arguments(use_numba, {True, False, "infer"})
#     if use_numba == "infer":
#         if "numba" in sys.modules:
#             use_numba = True
#         else:
#             use_numba = False
#         if verbose:
#             print("Numba is available:", use_numba, file=sys.stderr)

#     # Compute using numba
#     if use_numba:
#         assert "numba" in sys.modules
#         from numba import jit

#         def _biweight_midcorrelation_numba(A):
#             @jit
#             def _condensed_to_dense(n, m, est, norms, result):
#                 for i in range(m):
#                     for j in range(i + 1, m):
#                         x = 0
#                         for k in range(n):
#                             x += est[k, i] * est[k, j]
#                         result[i, j] = result[j, i] = x / norms[i] / norms[j]
#             n, m, est, norms = _base_computation(A)
#             result = np.empty((m, m))
#             np.fill_diagonal(result, 1.0)
#             _condensed_to_dense(n, m, est, norms, result)
#             return result

#         result = _biweight_midcorrelation_numba(X)
#     # Compute using numpy
#     else:
#         def _biweight_midcorrelation_numpy(A):
#             n, m, est, norms = _base_computation(A)
#             return np.einsum('mi,mj->ij', est, est) / norms[:, None] / norms[None, :]
#         result = _biweight_midcorrelation_numpy(X)

#     # Add labels
#     if labels is not None:
#         result = pd.DataFrame(result, index=labels, columns=labels)

#     return result



# ==============================================================================
# Modules
# ==============================================================================
# Built-ins
import os, sys, time, datetime, copy, pathlib
from collections import OrderedDict, defaultdict, Mapping
from io import StringIO, BytesIO
# from ast import literal_eval

# Compression & Serialization
import pickle, gzip, bz2, zipfile

# PyData
import pandas as pd

# Soothsayer
from soothsayer.utils import format_path, to_precision, add_objects_to_globals

# Soothsayer Utils
import soothsayer_utils as syu
functions_from_soothsayer_utils = [
    "read_dataframe",
    "write_dataframe",
    "read_from_clipboard",
]
add_objects_to_globals(syu, functions_from_soothsayer_utils, globals(), add_version=True, __all__=None)

# ======
# Future
# ======
# (1) Add ability for `path` arguments to be `pathlib.Path`
# (2) Add ability to process ~ in path
# ======
# Tables
# ======
# # Read dataframe
# def read_dataframe(path:str, sep="infer", index_col=0, header=0, compression="infer", pickled="infer", func_index=None, func_columns=None, evaluate_columns=None, engine="c", verbose=False, excel="infer", infer_series=False, sheet_name=None,  **args):

#     start_time = time.time()
#     path = format_path(path, str)
#     dir , ext = os.path.splitext(path)
#     ext = ext.lower()

#     if excel == "infer":
#         if ext in {".xlsx", ".xls"}:
#             excel = True
#         else:
#             excel = False

#     if excel:
#         if "sheetname" in args:
#             sheet_name = args.pop("sheetname")
#             print("DEPRECATED: Use `sheet_name` instead of `sheetname`", file=sys.stderr)
#         df = pd.read_excel(path, sheet_name=sheet_name, index_col=index_col, header=header, **args)

#     else:
#         # Seperator
#         if any(list(map(lambda x:ext.endswith(x),[".csv", "csv.gz", "csv.zip"]))):
#             sep = ","
#         else:
#             sep = "\t"

#         # Serialization
#         if pickled == "infer":
#             if ext in {".pkl", ".pgz", ".pbz2"}:
#                 pickled = True
#             else:
#                 pickled = False
#         # Compression
#         if compression == "infer":
#             if pickled:
#                 if ext == ".pkl":
#                     compression = None
#                 if ext == ".pgz":
#                     compression = "gzip"
#                 if ext == ".pbz2":
#                     compression = "bz2"
#             else:
#                 if ext == ".gz":
#                     compression = "gzip"
#                 if ext == ".bz2":
#                     compression = "bz2"
#                 if ext == ".zip":
#                     compression = "zip"


#         if pickled:
#             df = pd.read_pickle(path, compression=compression)
#         else:
#             df = pd.read_csv(path, sep=sep, index_col=index_col, header=header, compression=compression, engine=engine, **args)

#     condition_A = any([(excel == False), (sheet_name is not None)])
#     condition_B = all([(excel == True), (sheet_name is None)])

#     if condition_A:
#         # Map indices
#         if func_index is not None:
#             df.index = df.index.map(func_index)
#         if func_columns is not None:
#             df.columns = df.columns.map(func_columns)

#         if evaluate_columns is not None:
#             for field_column in evaluate_columns:
#                 try:
#                     df[field_column] = df[field_column].map(eval)
#                 except ValueError:
#                     if verbose:
#                         print(f"Could not use `eval` on column=`{field_column}`", file=sys.stderr)
#         if verbose:
#             print(f"{path.split('/')[-1]} | Dimensions: {df.shape} | Time: {to_precision(time.time() - start_time)} seconds", file=sys.stderr)
#     if condition_B:
#         if verbose:
#             print(f"{path.split('/')[-1]} | Sheets: {len(df)} | Time: {to_precision(time.time() - start_time)} seconds", file=sys.stderr)
#     if infer_series:
#         if df.shape[1] == 1:
#             df = df.iloc[:,0]
#     return df

# # Write dataframe
# def write_dataframe(data, path:str, sep="\t", compression="infer", pickled="infer",  excel="infer", create_directory_if_necessary=False, **args):
#     start_time = time.time()
#     path = format_path(path, str)
#     _ , ext = os.path.splitext(path)
#     dir, filename = os.path.split(path)
#     if not os.path.exists(dir):
#         dir = str(pathlib.Path(dir).absolute())
#         os.makedirs(dir, exist_ok=True)

#     # Excel
#     if excel == "infer":
#         if ext in {".xlsx", ".xls"}:
#             excel = True
#         else:
#             excel = False

#     if excel:
#         if not isinstance(data, Mapping):
#             data = {"Sheet1":data}
#         writer = pd.ExcelWriter(path)
#         for sheet_name, df in data.items():
#             df.to_excel(writer, sheet_name=str(sheet_name))
#         writer.save()
#     else:
#         # Serialization
#         if pickled == "infer":
#             if ext in {".pkl", ".pgz", ".pbz2"}:
#                 pickled = True
#             else:
#                 pickled = False
#         # Compression
#         if compression == "infer":
#             if pickled:
#                 if ext == ".pkl":
#                     compression = None
#                 if ext == ".pgz":
#                     compression = "gzip"
#                 if ext == ".pbz2":
#                     compression = "bz2"
#             else:
#                 compression = None
#                 if ext == ".gz":
#                     compression = "gzip"
#                 if ext == ".bz2":
#                     compression = "bz2"

#         if pickled:
#             data.to_pickle(path, compression=compression, **args)
#         else:
#             data.to_csv(path, sep=sep, compression=compression, **args)

# Read results from feature selection
def read_multiple_synopsis(dir:str, from_algorithm:str="clairvoyance"):
    """
    Assumes the following directory structure:
    __________________________________
    | project_directory
    | | synopsis_dir_for_all_submodels
    | | | {name}_synopsis
    | | | | {name}__synopsis.tsv
    ----------------------------------
    """
    accepted_args = ["clairvoyance"]
    def _read_clairvoyance(dir:str):
        # Get synopsis
        _data = list()
        for file in os.scandir(dir):
            if file.is_dir():
                submodel = file.name.split("_")[1]
                for file2 in os.scandir(file):
                    if file2.name.endswith("__synopsis.tsv"):
                        _df = read_dataframe(file2)
                        _df.insert(loc=0, column="submodel", value=submodel)
                        _data.append(_df)
        df_synopsis = pd.concat(_data, axis=0).reset_index()

        # Evaluate fields
        fields_to_evaluate = ["hyperparameters"] + [*filter(lambda field:field.endswith("set"), df_synopsis.columns)]
        for field in fields_to_evaluate:
            df_synopsis[field] = df_synopsis[field].map(eval)
        return df_synopsis

    assert from_algorithm in accepted_args, f"{from_algorithm} is not in {accepted_args}"
    dir = format_path(dir, str)
    if from_algorithm == "clairvoyance":
        return _read_clairvoyance(dir)



# ==============================================================================
# Imports
# ==============================================================================
# Built-ins
import os, sys, time

import pandas as pd
import numpy as np

# Soothsayer
from ..r_wrappers import *
from soothsayer.utils import check_packages

# ==============================================================================
# R Imports
# ==============================================================================
if "rpy2" in sys.modules:
    from rpy2 import robjects as ro
    from rpy2 import rinterface as ri

    from rpy2.robjects.packages import importr
    try:
        from rpy2.rinterface import RRuntimeError
    except ImportError:
        from rpy2.rinterface_lib.embedded import RRuntimeError
    from rpy2.robjects import pandas2ri
    # pandas2ri.activate()
    R = ro.r
    NULL = ri.NULL
    #rinterface.set_writeconsole_regular(None)

# R packages
# WGCNA = R_package_retrieve("WGCNA")
#
# def TOMsimilarity(adjacency,  TOMType="unsigned",  TOMDenom="min"):
#     """
#     WGCNA: TOMsimilarity
#         function (adjMat, TOMType = "unsigned", TOMDenom = "min", verbose = 1,
#                    indent = 0)
#     """
#     assert np.all(adjacency.index == adjacency.columns), "pd.DataFrame must have symmetric labels for index and columns"
#     nodes = adjacency.index
#     rDF_adj = pandas_to_rpy2(adjacency)
#     rDF_tom = wgcna.TOMsimilarity(R["as.matrix"](rDF_adj), TOMType=TOMType, TOMDenom=TOMDenom, verbose=0)
#     Ar_tom = rpy2_to_pandas(rDF_tom)
#     return pd.DataFrame(Ar_tom, index=nodes, columns=nodes)
#
# def bicor(X, axis=1, n_jobs=-1):
#     """
#     WGCNA: bicor
#         function (x, y = NULL, robustX = TRUE, robustY = TRUE, use = "all.obs",
#                    maxPOutliers = 1, qu <...> dian absolute deviation, or zero variance."))
#     """
#     if n_jobs == -1:
#         n_jobs = multiprocessing.cpu_count()
#
#     if axis == 1: X_copy = X.copy()
#     if axis == 0: X_copy = X.copy().T
#     labels = X_copy.columns
#     rDF_sim = WGCNA.bicor(pandas_to_rpy2(X_copy))
#     df_bicor = pd.DataFrame(rpy2_to_pandas(rDF_sim), index=labels, columns=labels)
#     return df_bicor

@check_packages(["WGCNA"], language="r", import_into_backend=False)
def pickSoftThreshold_fromSimilarity(df_adj, query_powers):
    WGCNA = R_package_retrieve("WGCNA")

    # Run pickSoftThreshold.fromSimilarity
    query_powers = ro.IntVector(list(query_powers))
    r_adj = pandas_to_rpy2(df_adj)
    rDF_scalefreetopology = WGCNA.pickSoftThreshold_fromSimilarity(R["as.matrix"](r_adj), powerVector = query_powers, verbose=0, moreNetworkConcepts=True)[1]
    return rpy2_to_pandas(rDF_scalefreetopology)



# ==============================================================================
# Modules
# ==============================================================================
# Built-ins
import os, sys, time, datetime, pathlib
from collections import OrderedDict, defaultdict
from io import StringIO, BytesIO, TextIOWrapper
from ast import literal_eval

# Compression & Serialization
import pickle, gzip, bz2, zipfile

# PyData
import pandas as pd

# Soothsayer
from soothsayer.utils import format_path, infer_compression, is_file_like, add_objects_to_globals

# Soothsayer Utils
import soothsayer_utils as syu
functions_from_soothsayer_utils = [
    "get_file_object",
    "read_textfile",
    "read_from_clipboard",
]
add_objects_to_globals(syu, functions_from_soothsayer_utils, globals(), add_version=True, __all__=None)


# # Create file object
# def get_file_object(path:str, mode="infer", compression="infer", safe_mode="infer", verbose=True):
#     """
#     # Should this be in io?
#     with get_file_object("./test.txt.zip", mode="infer", verbose=False) as f_read:
#     with get_file_object("./test_write.txt.bz2", mode="write", verbose=False) as f_write:
#         for line in f_read.readlines():
#             line = str(line.strip())
#             print(line, file=f_write)
#     """
#     # Init
#     f = None
#     file_will_be_written = False

#     # Paths
#     path = format_path(path)
#     path_exists = os.path.exists(path)

#     if verbose:
#         header = " ".join(["Processing file:", path])
#         print("="*len(header), header, "="*len(header), sep="\n", file=sys.stderr)
#     if compression == "infer":
#         compression = infer_compression(path)
#         if verbose:
#             print("Inferring compression:", compression, file=sys.stderr)

#     # Inferring mode
#     if mode == "infer": # Create new function for this? infer_filemode?
#         if path_exists:
#             mode = "read"
#         else:
#             mode = "write"
#     assert mode != "infer", "The mode should be inferred by this point.  Please specify mode manually."
#     assert compression != "infer", "The compression should be inferred by this point.  Please specify compression manually."

#     # Generic read write
#     if mode in ["write", "read"]:
#         if mode == "write":
#             mode = "w"
#         if mode == "read":
#             mode = "r"
#         if compression in ["gzip", "bz2"]:
#             mode = mode + "b"
#         if verbose:
#             print("Inferring mode:", mode, file=sys.stderr)

#     # Will a file be written?
#     if "w" in mode:
#         file_will_be_written = True

#     # Ensure zip is not being written
#     if file_will_be_written:
#         assert compression != "zip", "Currently cannot handle writing zip files.  Please use gzip, bz2, or None."
#         # Future do this:
#         # https://docs.python.org/3/library/zipfile.html#zipfile.ZipFile.open

#     # Safe mode
#     if safe_mode == "infer":
#         if file_will_be_written:
#             safe_mode = True
#         else:
#             safe_mode = False
#     assert safe_mode in {True,False}, "Please choose either True or False for `safe_mode`"
#     if safe_mode:
#         if all([file_will_be_written, path_exists]):
#             raise Exception(f"Safe Mode: Please explicitly provide a writeable mode ['w', 'wb', or 'write'] because `{path}` already exists and will be rewritten.")

#     # GZIP compression
#     if compression == "gzip":
#         f = gzip.open(path, mode)
#     # BZ2 compression
#     if compression == "bz2":
#         f = bz2.open(path, mode)
#     if compression == "zip":
#         filename, ext = os.path.splitext(os.path.split(path)[-1])
#         f = zipfile.ZipFile(path,mode).open(filename)
#     # No compression
#     if f is None:
#         return open(path, mode)

#     # Reading and writing compressed files
#     else:
#         return TextIOWrapper(f, encoding="utf-8")


# # Text reading wrapper
# def read_textfile(path, enum=True, generator=True, mode="read", compression="infer", into=pd.Series):
#     """
#     2018-May-29
#     Edits: 2018-December-27: Added `get_file_object` dependency
#     """
#     assert mode not in ["w", "wb", "a"], "`mode` should not be in {w, wb, a} because it will overwrite"
#     assert compression  in ["infer", "gzip", "bz2", "zip", None], "Valid `compression` types are 'infer', 'gzip', 'bz2', 'zip'"
#     # Format file path
#     path = format_path(path, str)
#     # Get file object
#     handle = get_file_object(path=path, mode=mode, compression=compression, safe_mode=False, verbose=False)
#     # Subfunctions
#     def run_return_list(handle):
#         data = handle.read().split("\n")
#         handle.close()
#         if into == pd.core.series.Series:
#             return pd.Series(data, name=path)
#         else:
#             if enum:
#                 return into([*enumerate(data)])
#             if not enum:
#                 return into(data)

#     def run_return_iterable(handle):
#         if enum:
#             for i,line in enumerate(handle.readlines()):
#                 yield i, line.strip()
#         if not enum:
#             for line in handle.readlines():
#                 yield line.strip()
#         handle.close()

#     # Controller
#     if generator:
#         return run_return_iterable(handle=handle)
#     if not generator:
#         return run_return_list(handle=handle)



# ==============================================================================
# Modules
# ==============================================================================
# Built-ins
import os, sys, time, pathlib, warnings, requests
from collections import *
from io import StringIO, BytesIO
from ast import literal_eval
from tqdm import tqdm
import xml.etree.ElementTree as ET

# Compression & Serialization
import pickle, gzip, bz2, zipfile

# PyData
import pandas as pd
import numpy as np

# Biology
from Bio import SeqIO

# Soothsayer
from soothsayer.utils import format_path, infer_compression, is_path_like, is_query_class, assert_acceptable_arguments, add_objects_to_globals
from ..text import read_textfile, get_file_object

# Soothsayer Utils
import soothsayer_utils as syu
functions_from_soothsayer_utils = [
'read_fasta',
"read_star_log",
'read_blast',
'read_hmmer',
'read_gtf_gff_base',
'read_gff3', 
'read_gtf', 
'read_ncbi_xml', 
'write_fasta', 
]
add_objects_to_globals(syu, functions_from_soothsayer_utils, globals(), add_version=True, __all__=None)

# ======
# Future
# ======
# (1) Add ability for `path` arguments to be `pathlib.Path`
# (2) Add ability to process ~ in path



# ==============
# Bioinformatics
# ==============
# Clustalo Distance Matrix
def read_clustalo_distmat(path:str):
    path = format_path(path)
    with open(path, "r") as f:
        for i, line in enumerate(f):
            line = line.strip()
            conditions = [
                line.startswith("#"),
            ]
            try:
                int(line)
                conditions.append(True)
            except ValueError:
                pass
            if not any(conditions):
                break
    df_tmp = pd.read_csv(path, sep="\t", index_col=0, header=None, skiprows=i)
    tmp = df_tmp.index.map(lambda line:[x for x in line.split(" ") if len(x) > 0])
    if ":" in tmp[0][0]:
        tmp = tmp.map(lambda x:x[1:])
    data = OrderedDict()
    try:
        for line in tmp:
            data[line[0]] = pd.Series([*map(float,line[1:])], name=line[0])
        df_dism = pd.DataFrame(data)
        df_dism.index = df_dism.columns
        return df_dism
    except ValueError:
        print("Could not accurately parse clustal distance matrix.  Returning raw data instead.", file=sys.stderr)
        return tmp

# VizBin corrdinates
def read_vizbin_coords(xy_path:str, fasta_path:str=None, xlabel="x", ylabel="y", write_coords=True):
    """
    VizBin: http://claczny.github.io/VizBin/
    xy_path: VizBin coordinates file path
    fasta_path: Input fasta file path

    Usage:
    xy_path = "/var/folders/6z/5vbtz_gmkr76ftgc3149dvtr0003c0/T/map1723096415078571924/points.txt"
    fasta_path = "./assembly.fa")
    df_xy = read_vizbin_coords(xy_path, fasta_path)
    """
    xy_path = format_path(xy_path)

    df_xy = pd.read_csv(xy_path, sep=",", header=None).astype(float)
    path_write = None
    if fasta_path is not None:
        fasta_path = format_path(fasta_path)
        df_xy.index = [*map(lambda record:str(record.id), SeqIO.parse(fasta_path, "fasta"))]
        if write_coords == True:
            path_write = f"{fasta_path}.xy"
    df_xy.columns = [xlabel,ylabel]
    if path_write is not None:
        df_xy.to_csv(path_write, sep="\t")
    return df_xy

# CheckM QA
def read_checkm_qa(path:str, format=8, verbose=True, rename_columns=True, func_orf_to_contig=lambda id_orf:"_".join(id_orf.split("_")[:-1])):
    """
    Tested for v1.0.7
    Read CheckM qa output into pd.DataFrame
    accepted_formats = {2,8}
    """
    accepted_formats = {2,8}
    assert format in accepted_formats, f"Currently only supporting formats: `{accepted_formats}`"
    path = format_path(path)

    # QA: Format = 2
    def _read_format_qa_2(path, rename_columns, verbose):
        d_rename = {
            "Bin Id":"id_bin",
            "Marker lineage":"marker_lineage",
            "# genomes":"number_of_genomes",
            "# markers":"number_of_markers",
            "# marker sets":"number_of_marker-sets",
            "Completeness":"completeness",
            "Contamination":"contamination",
            "Strain heterogeneity":"strain_heterogeneity",
            "Genome size (bp)":"genome_size[bp]",
            "# ambiguous bases":"number_of_ambiguous-bases",
            "# scaffolds":"number_of_scaffolds",
            "# contigs":"number_of_contigs",
            "N50 (scaffolds)":"N50[scaffolds]",
            "N50 (contigs)":"N50[contigs]",
            "Mean scaffold length (bp)":"Mean_scaffold_length[bp]",
            "Mean contig length (bp)":"Mean_contig_length[bp]",
            "Longest scaffold (bp)":"Longest_scaffold[bp]",
            "Longest contig (bp)":"Longest_contig[bp]",
            "GC std (scaffolds > 1kbp)":"GC_std[scaffolds>1kbp]",
            "Coding density":"coding_density",
            "Translation table":"trans_table",
            "# predicted genes":"number_of_predicted-genes",
        }
        tab_table = False
        for i, line in read_textfile(path):
            if i == 0:
                if not line.startswith("-"):
                    tab_table = True
                    break
            if i == 1:
                columns = pd.Index([*map(lambda x: x.strip(), filter(bool, line.split("  ")))])
                break
        if tab_table:
            df_table = pd.read_csv(path, sep="\t", index_col=None)
            columns = df_table.columns
        else:
            table = pd.read_csv(path, sep="\t", skiprows=[0,1,2],skipfooter=1, header=None).values.ravel()
            df_table = pd.DataFrame([*map(lambda line: [*filter(bool, line.split(" "))], table)])
            df_table.columns = columns
        if verbose:
            print("Total bins:", df_table.shape[0], file=sys.stderr)
        if rename_columns:
            df_table.columns = columns.map(lambda c: d_rename[c] if c in d_rename else c)
            return df_table.set_index("id_bin", drop=True).sort_values(["completeness", "contamination"], ascending=[False, True])
        else:
            return df_table.set_index("Bin Id", drop=True).sort_values(["Completeness", "Contamination"], ascending=[False, True])
    # QA: Format = 8
    def _read_format_qa_8(path, verbose):
        table = list()
        for i,line in read_textfile(path):
            if i > 0:
                data = line.replace("\n","").split("\t")
                # Get bin ID
                ids = data[:2]
                # Get HMM IDs
                hmm_ids = set([x.split(".")[0].split(",")[0] for x in data[2:]  ])
                # Concat positions
                if i == 0:
                    hmm_pos = data[2:]
                else:
                    hmm_pos = set([*map(lambda organized_hit: (organized_hit[0], int(organized_hit[1]), int(organized_hit[2])), map(lambda hit: tuple(hit.split(",")), data[2:]))])
                # Create table
                line_out = ids + [hmm_ids] + [hmm_pos]
                table.append(line_out)
        # Build table
        df_table = pd.DataFrame(table)
        df_table.columns = ["id_bin", "id_orf", "markers", "marker_positions"]

        # Split ORFs
        table = list()
        for i, row in df_table.iterrows():
            if "&&" in row["id_orf"]:
                for id_orf in row["id_orf"].split("&&"):
                    sub_row = row.copy()
                    sub_row["id_orf"] = id_orf
                    table.append(sub_row)
            else:
                table.append(row)
        df_table = pd.DataFrame(table)

        if func_orf_to_contig is not None:
            df_table.insert(loc=1, column="id_contig", value=df_table["id_orf"].map(func_orf_to_contig).tolist())
        if verbose:
            print("Total markers:",len(set().union(*df_table["markers"])), sep="\t", file=sys.stderr)
        return df_table

    # Controller
    if format == 2:
        return _read_format_qa_2(path, rename_columns, verbose)
    if format == 8:
        return _read_format_qa_8(path, verbose)

# # # Read Fasta File
# # def read_fasta(path:str, description:bool=True, case="upper", func_header=None, into=pd.Series, compression="infer", name=None, verbose=True):
# #     """
# #     Reads in a single fasta file or a directory of fasta files into a dictionary.
# #     """
# #     # Get path
# #     path = format_path(path)

# #     # Assign pathname as name if there isn't one
# #     if name is None:
# #         name = path

# #     # Open file object
# #     f = get_file_object(path, mode="read", compression=compression, safe_mode=False, verbose=False)

# #     # Read in fasta
# #     d_id_seq = OrderedDict()

# #     if verbose:
# #         seq_records = tqdm(SeqIO.FastaIO.SimpleFastaParser(f), f"Reading sequence file: {path}")
# #     else:
# #         seq_records = SeqIO.FastaIO.SimpleFastaParser(f)

# #     # Verbose but faster
# #     if case == "lower":
# #         for header, seq in seq_records:
# #             seq = seq.lower()
# #             if not description:
# #                 header = header.split(" ")[0]
# #             d_id_seq[header] = seq
# #     if case == "upper":
# #         for header, seq in seq_records:
# #             seq = seq.upper()
# #             if not description:
# #                 header = header.split(" ")[0]
# #             d_id_seq[header] = seq
# #     if case is None:
# #         for header, seq in seq_records:
# #             if not description:
# #                 header = header.split(" ")[0]
# #             d_id_seq[header] = seq

# #     # Close File
# #     f.close()

# #     # Transform header
# #     if func_header is not None:
# #         d_id_seq = OrderedDict( [(func_header(id),seq) for id, seq in d_id_seq.items()])
# #     sequences = into(d_id_seq)
# #     if hasattr(sequences, "name"):
# #         sequences.name = name
# #     return sequences


# # # Writing sequence files
# # def write_fasta(sequences, path:str, compression="infer"):
# #     """
# #     Sequence stats:
# #     count    29999.000000
# #     mean       310.621754
# #     std       1339.422833
# #     min         56.000000
# #     25%         75.000000
# #     50%        111.000000
# #     75%        219.000000
# #     max      54446.000000

# #     Benchmarks:
# #     No compression: 47.2 ms ± 616 µs per loop (mean ± std. dev. of 7 runs, 10 loops each)
# #     Gzip: 9.85 s ± 261 ms per loop (mean ± std. dev. of 7 runs, 1 loop each)
# #     Bzip2: 864 ms ± 16.1 ms per loop (mean ± std. dev. of 7 runs, 1 loop each)
# #     """
# #     # path = format_path(path)
# #     # if compression == "infer":
# #     #     compression = infer_compression(path)
# #     if is_query_class(path, ["stdout", "stderr", "streamwrapper"]):
# #         path.writelines(f">{id}\n{seq}\n" for id, seq in sequences.items())
# #     else:
# #         with get_file_object(path, mode="write", compression=compression, safe_mode=False, verbose=False) as f:
# #             f.writelines(f">{id}\n{seq}\n" for id, seq in sequences.items())

# # Read blast output
# def read_blast(path:str, length_query=None, length_subject=None, sort_by="bitscore"):
#     """
#     if 12: ['qseqid', 'sseqid', 'pident', 'length', 'mismatch', 'gapopen', 'qstart', 'qend', 'sstart', 'send', 'evalue', 'bitscore']
#     if 15 assumes: -outfmt '6 std qlen slen stitle': ["std", "qlen", "slen", "stitle"]
#     ####################################################
#     Column    NCBI name    Description
#     1    qseqid    Query Seq-id (ID of your sequence)
#     2    sseqid    Subject Seq-id (ID of the database hit)
#     3    pident    Percentage of identical matches
#     4    length    Alignment length
#     5    mismatch    Number of mismatches
#     6    gapopen    Number of gap openings
#     7    qstart    Start of alignment in query
#     8    qend    End of alignment in query
#     9    sstart    Start of alignment in subject (database hit)
#     10    send    End of alignment in subject (database hit)
#     11    evalue    Expectation value (E-value)
#     12    bitscore    Bit score
#     13    sallseqid    All subject Seq-id(s), separated by a ';'
#     14    score    Raw score
#     15    nident    Number of identical matches
#     16    positive    Number of positive-scoring matches
#     17    gaps    Total number of gaps
#     18    ppos    Percentage of positive-scoring matches
#     19    qframe    Query frame
#     20    sframe    Subject frame
#     21    qseq    Aligned part of query sequence
#     22    sseq    Aligned part of subject sequence
#     23    qlen    Query sequence length
#     24    slen    Subject sequence length
#     25    salltitles    All subject title(s), separated by a '<>'

#     Example inputs:
#         * blat -prot Yeast/Saccharomyces_cerevisiae.R64-1-1.pep.all.processed.fa  Phaeodactylum_tricornutum.ASM15095v2.pep.all.processed.fa -out=blast8 yeast-pt.blast8
#         * diamond blastp -d ../../../../reference_data/references/gracilibacteria/reference_proteins.nmnd -q ./prodigal_output/orfs.faa -f 6 -o ./diamond_output/output.blast6

#     """
#     path = format_path(path)
#     idx_default_fields = ['qseqid', 'sseqid', 'pident', 'length', 'mismatch', 'gapopen', 'qstart', 'qend', 'sstart', 'send', 'evalue', 'bitscore']
#     # ["query_id", "subject_id", "identity", "alignment_length", "mismatches", "gap_openings", "query_start", "query_end", "subject_start", "subject_end", "e-value", "bit_score"]
#     df_blast = pd.read_csv(path, header=None, sep="\t")
#     if df_blast.shape[1] == 12:
#         df_blast.columns = idx_default_fields
#     if df_blast.shape[1] == 15:
#         df_blast.columns = idx_default_fields + ["qlen", "slen", "stitle"]
#     # Length of query
#     if length_query is not None:
#         if is_path_like(length_query):
#             length_query = read_fasta(length_query, description=False, verbose=False)
#         if isinstance(length_query[0], str):
#             length_query = length_query.map(len)
#         df_blast["qlength"] = df_blast["qseqid"].map(lambda id: length_query[id])
#         df_blast["qratio"] = (df_blast["qend"] - df_blast["qstart"])/df_blast["qlength"]
#     # Length of subject
#     if length_subject is not None:
#         if is_path_like(length_subject):
#             length_subject = read_fasta(length_subject, description=False, verbose=False)
#         if isinstance(length_subject[0], str):
#             length_subject = length_subject.map(len)
#         df_blast["slength"] = df_blast["sseqid"].map(lambda id: length_subject[id])
#         df_blast["sratio"] = (df_blast["send"] - df_blast["sstart"])/df_blast["slength"]
#     if sort_by is not None:
#         df_blast = df_blast.sort_values(by=sort_by, ascending=False).reset_index(drop=True)
#     return df_blast

#     return df_blast
# # Helper function for reading gtf and gff3
# def _read_gtf_gff_base(path, compression, record_type, verbose):
#     # Read the gff3 file
#     with get_file_object(path, mode="read", compression=compression, safe_mode=False, verbose=False) as f:
#         if verbose:
#             iterable_lines = tqdm(f.readlines(), "Processing lines")
#         else:
#             iterable_lines = f.readlines()
#         data= list()
#         if record_type is None:
#             for line in iterable_lines:
#                 if not line.startswith("#"):
#                     line = line.strip("\n")
#                     if bool(line):
#                         base_fields = line.split("\t")
#                         data.append(base_fields)
#         else:
#             for line in iterable_lines:
#                 if not line.startswith("#"):
#                     if f"{record_type}_id" in line:
#                         line = line.strip("\n")
#                         base_fields = line.split("\t")
#                         data.append(base_fields)
#     # Generate table
#     df_base = pd.DataFrame(data)
#     df_base.columns = ["seq_record", "source", "seq_type", "pos_start", "pos_end", ".1", "sense", ".2", "data_fields"]
#     return df_base

# def read_gff3(path:str, compression="infer", record_type=None, verbose = True, reset_index=False, name=True):
#     def f(x):
#         fields = x.split(";")
#         data = dict()
#         for item in fields:
#             k, v = item.split("=")
#             data[k] = v
#         return data
#     path = format_path(path)
#     if verbose:
#         print("Reading gff3 file:",path,sep="\t", file=sys.stderr)
#     accepted_recordtypes = {"exon", "gene", "transcript", "protein", None}
#     assert record_type in accepted_recordtypes, "Unrecognized record_type.  Please choose from the following: {}".format(accepted_recordtypes)

#     # Read the gff3 file
#     df_base = _read_gtf_gff_base(path, compression, record_type, verbose)

#     try:
#         df_fields = pd.DataFrame(df_base["data_fields"].map(f).to_dict()).T
#         df_gff3 = pd.concat([df_base[["seq_record", "source", "seq_type", "pos_start", "pos_end", ".1", "sense", ".2"]], df_fields], axis=1)
#          # Contains identifier
#         if "ID" in df_gff3.columns:
#             df_gff3["seq_id"] = df_gff3["ID"].map(lambda x: "-".join(x.split("-")[1:]) if not pd.isnull(x) else x)
#         if reset_index:
#             df_gff3 = df_gff3.reset_index(drop=True)
#     except IndexError:
#         warnings.warn("Could not successfully parse file: {}".format(path))
#         df_gff3 = df_base

#     # Path index
#     if name is not None:
#         if name == True:
#             name = path
#         df_gff3.index.name = name

#     return df_gff3

# # GTF
# def read_gtf(path:str, compression="infer", record_type=None, verbose = True, reset_index=False, name=True):
#     path = format_path(path)
#     if verbose:
#         print(f"Reading gtf file:",path,sep="\t", file=sys.stderr)
#     accepted_recordtypes = {"exon", "gene", "transcript", "protein", None}
#     assert record_type in accepted_recordtypes, f"Unrecognized record_type.  Please choose from the following: {accepted_recordtypes}"

#     # Read the gtf file
#     df_base = _read_gtf_gff_base(path, compression, record_type, verbose)

#     # Splitting fields
#     iterable_fields =  df_base.iloc[:,-1].iteritems()

#     # Parse fields
#     dataframe_build = dict()
#     for i, gtf_data in iterable_fields:
#         gtf_data = gtf_data.replace("''","").replace('"',"")
#         fields = filter(bool, map(lambda field:field.strip(), gtf_data.split(";")))
#         fields = map(lambda field: field.split(" "), fields)
#         dataframe_build[i] = dict(fields)
#     df_fields = pd.DataFrame(dataframe_build).T
#     df_gtf = pd.concat([df_base.iloc[:,:7], df_fields], axis=1)

#     # Reset index
#     if reset_index:
#         df_gtf = df_gtf.reset_index(drop=True)

#     # Path index
#     if name is not None:
#         if name == True:
#             name = path
#         df_gtf.index.name = name
#     return df_gtf

# # NCBI XML
# def read_ncbi_xml(path:str, index_name=None):
#     # Format path
#     path = format_path(path)
#     # Parse the XML tree
#     tree = ET.parse(path)
#     root = tree.getroot()
#     # Extract the attributes
#     data = defaultdict(dict)
#     for record in tqdm(root.getchildren(), "Reading NCBI XML: {}".format(path)):
#         id_record = record.attrib["accession"]
#         for attribute in record.findall("Attributes/*"):
#             data[id_record][attribute.attrib["attribute_name"]] = attribute.text
#     # Create pd.DataFrame
#     df = pd.DataFrame(data).T
#     df.index.name = index_name
#     return df

# Read EMBL-EBI sample metadata
def read_ebi_sample_metadata(query, base_url="https://www.ebi.ac.uk/metagenomics/api/v1/samples/{}/metadata", mode="infer"):
    """
    query:
        id_sample: "ERS488919"
        url: "https://www.ebi.ac.uk/metagenomics/api/v1/samples/ERS488919/metadata
        file: "/path/to/ERS488919.json"

    JSON file can be retrieved via:
        curl -X GET "https://www.ebi.ac.uk/metagenomics/api/v1/samples/ERS488919/metadata" -H "accept: application/json" > ERS488919.json
    """
    # Acceptable arguments for mode
    assert_acceptable_arguments(mode, {"infer", "cloud", "local"})

    # Functions
    def _get_url(query, base_url):
        # Is a URL provided?
        if any(map(lambda x: x in query, ["http", "www"])):
            url = query
        # Is a sample identifier provided
        else:
            url = base_url.format(query)
        return url

    # Infer mode
    url = None
    response = None
    if mode == "infer":
        try:
            url = _get_url(query, base_url)
            response = requests.get(url)
            mode = "cloud"
        except requests.exceptions.MissingSchema:
            mode = "local"

    # Cloud
    if mode == "cloud":
        if response is None:
            # Get cloud data
            url = _get_url(query, base_url)
            response = requests.get(url)
        data = response.json()["data"]
    if mode == "local":
        with open(query, "r") as f:
            data = json.loads(f.read())["data"]
    # Determine sample identifier
    id_sample = data[0]["relationships"]["sample"]["data"]["id"]

    return pd.Series(
            data=dict(map(lambda item: (item["attributes"]["key"], item["attributes"]["value"]), data)),
            name=id_sample,
    )



# ==============================================================================
# Modules
# ==============================================================================
# Built-ins
import os, sys, time, datetime, logging, uuid, pathlib, posix, gzip, bz2, zipfile, subprocess, requests, operator
from collections import OrderedDict, defaultdict, Mapping
from io import TextIOWrapper

# PyData
import pandas as pd
import numpy as np
import seaborn as sns
from scipy.interpolate import griddata

## Matplotlib
import matplotlib.pyplot as plt
from matplotlib.colors import rgb2hex, hex2color, to_rgb, to_rgba #, ListedColormap
from palettable.cartocolors.diverging import TealRose_3
from palettable.cmocean.sequential import Tempo_8

# Datasets
from sklearn.datasets import load_iris

# Soothsayer
import soothsayer_utils as syu

functions_from_soothsayer_utils = [
'assert_acceptable_arguments', 'boolean', 'consecutive_replace', 'contains', 'dict_build', 'dict_collapse', 'dict_expand', 'dict_fill', 'dict_filter', 'dict_reverse', 'dict_tree',"dict_py_to_bash",
'flatten', 'format_memory', 'format_duration', 'format_header', 'format_path', 'fragment',  'get_timestamp', 'get_unique_identifier', 'hash_kmer', 'infer_compression', 'is_all_same_type',
'is_dict', 'is_dict_like', 'is_file_like', 'is_function', 'is_in_namespace', 'is_nonstring_iterable', 'is_symmetrical', 'is_number', 'is_path_like', 'is_query_class', "is_graph", "is_color",
'iterable_depth', 'join_as_strings',
'pad_left', 'pv', 'range_like',  'reverse_complement', 'to_precision', "check_packages", "Suppress","get_color_midpoint","remove_pairwise_nan", "format_pairwise",
]

__all__ = {
    # Tree
    "infer_tree_type", "check_polytomy", "is_leaf","name_tree_nodes","prune_tree",
    # Need to organize these
    "pd_prepend_level_to_index", 'pd_series_to_groupby_to_dataframe', 'pd_dropduplicates_index', 'generate_random_sequence', 'is_color', 'add_cbar_from_data', 'get_coords_contour', 'get_repr', 'get_parameters_ellipse', 'DIVERGING_KWS', 'determine_mode_for_logfiles', 'create_logfile', 'is_symmetrical', 'pd_dataframe_matmul', 'dataframe_to_matrixstring', 'infer_cmap', 'pd_dataframe_extend_index', 'configure_scatter', 'CMAP_DIVERGING', 'pd_series_collapse', 'is_graph', 'format_filename', 'pd_series_filter', 'map_colors', 'rgb_to_rgba', 'LEGEND_KWS', 'force_symmetry', 'is_rgb_like', 'COLOR_POSITIVE', 'Chromatic', 'COLOR_NEGATIVE', 'infer_vmin_vmax', 'pd_dataframe_query', 'get_coords_centroid', 'filter_compositional', 'scalarmapping_from_data', 'infer_continuous_type', 
    'format_mpl_legend_handles', "add_objects_to_globals",
    }



# Add objects to globals
def add_objects_to_globals(module, objects:set, globals, add_version:bool=True, __all__:set=None):
    for obj_name in objects:
        obj = getattr(module, obj_name)
        if add_version:
            version = "{}({})".format(module.__name__, module.__version__)
            setattr(obj, "__version__", version)
        globals[obj_name] = obj
        if __all__ is not None:
            assert isinstance(__all__, set)
            __all__.add(obj_name)
            
add_objects_to_globals(syu, functions_from_soothsayer_utils, globals(), add_version=True, __all__=__all__)

__all__ = sorted(__all__)


# =============
# Defaults
# =============
LEGEND_KWS = {'fontsize': 15, 'frameon': True, 'facecolor': 'white', 'edgecolor': 'black', 'loc': 'center left', 'bbox_to_anchor': (1, 0.5)}
DIVERGING_KWS = dict(h_neg=220, h_pos=15, sep=20, s=90, l=50)
CMAP_DIVERGING = sns.diverging_palette(**DIVERGING_KWS, as_cmap=True)
COLOR_NEGATIVE, COLOR_POSITIVE = sns.diverging_palette(**DIVERGING_KWS, n=2).as_hex()

# ===========
# # Types
# # ===========
def is_rgb_like(c):
    condition_1 = type(c) != str
    condition_2 = len(c) in [3,4]
    return all([condition_1, condition_2])

# def is_color(obj):
#     # Note: This can't handle values that are RGB in (0-255) only (0,1)
#     try:
#         to_rgb(obj)
#         return True
#     except ValueError:
#         verdict = False
#         if is_nonstring_iterable(obj):
#             if all(type(x) in [float, int] for x in obj):
#                 if all(0 <= x <= 255 for x in obj):
#                     verdict = True
#         return verdict
# def is_graph(obj):
#     return hasattr(obj, "has_edge")

# =============
# Checks
# =============
# def is_symmetrical(X, tol=None):
#     if X.shape[0] == X.shape[1]:
#         if tol is None:
#             return np.all(np.tril(X) == np.triu(X).T)
#         if tol:
#             return (np.tril(X) - np.triu(X).T).ravel().min() < tol
#     else:
#         return False

def force_symmetry(X):
    return (X + X.T)/2

# =============
# Tree
# =============


# Tree
def infer_tree_type(tree):
    tree_type = None
    query_type = str(tree.__class__).split("'")[1].split(".")[0]
    if query_type in {"skbio"}:
        tree_type = "skbio"
    if query_type in {"ete3"}:
        tree_type = "ete"
    assert tree_type is not None, "Please use either skbio or ete3 tree.  Tree type deterined as {}".format(query_type)
    return tree_type

# Is node a leaf?
def is_leaf(node, tree_type="infer"):
    assert_acceptable_arguments(tree_type, {"ete", "infer", "skbio"})
    if tree_type == "infer":
        tree_type = infer_tree_type(node)
    function_name = {"skbio":"is_tip", "ete":"is_leaf"}[tree_type]
    return getattr(node, function_name)()

# Check polytomy
def check_polytomy(tree, tree_type="infer"):
    assert_acceptable_arguments(tree_type, {"ete",  "infer", "skbio"})
    if tree_type == "infer":
        tree_type = infer_tree_type(tree)
        
    if tree_type == "ete":
        # Check bifurcation
        n_internal_nodes = len(list(filter(lambda node:node.is_leaf() == False, tree.traverse())))
        n_leaves = len(list(filter(lambda node:node.is_leaf(), tree.traverse())))
        if n_internal_nodes < (n_leaves - 1):
            raise Exception("Please resolve tree polytomy and force bifurcation: Use `tree.resolve_polytomy()` before naming nodes for `ete`")

    if tree_type == "skbio":
        # Check bifurcation
        n_internal_nodes = len(list(filter(lambda node:node.is_tip() == False, tree.traverse())))
        n_leaves = len(list(filter(lambda node:node.is_tip(), tree.traverse())))
        if n_internal_nodes < (n_leaves - 1):
            raise Exception("Please resolve tree polytomy and force bifurcation: Use `tree.bifurcate()` before naming nodes for `skbio`")


# Name tree nodes
def name_tree_nodes(tree,  node_prefix:str="y", tree_type:str="infer", polytomy_ok=False, inplace=False):
    """
    tree can be skbio or any ete3 tree
    """
    assert_acceptable_arguments(tree_type, {"ete", "infer", "skbio"})

    # In place or on a copy
    if not inplace:
        tree = tree.copy()
    # Determine tree type
    if tree_type == "infer":
        tree_type = infer_tree_type(tree)

    # Polytomy
    if not polytomy_ok:
        check_polytomy(tree=tree, tree_type=tree_type)
    # Name tree nodes
    intermediate_node_index = 1
    for node in tree.traverse():
        if not is_leaf(node):
            node.name = f"{node_prefix}{intermediate_node_index}"
            intermediate_node_index += 1
    if not inplace:
        return tree
    
# Prune tree
def prune_tree(tree,  leaves, tree_type="infer"):
    assert_acceptable_arguments(tree_type, {"ete",  "infer", "skbio"})
    if tree_type == "infer":
        tree_type = infer_tree_type(tree)
    if tree_type == "ete":
        tree.prune(leaves)
    if tree_type == "skbio":
        tree = tree.shear(leaves)
        tree.prune()
    return tree

# =============
# Miscellaneous
# =============
# Random sequence generation
def generate_random_sequence(size:int=100, alphabet=["A","T", "C", "G"], weights=[0.25,0.25,0.25,0.25], random_state=None):
    rs = np.random.RandomState(random_state)
    x = np.random.choice(alphabet, size=size, replace=True, p=weights)
    return "".join(x)

# For ete3 ClusterNode
def dataframe_to_matrixstring(df):
    return df.to_csv(None, sep="\t",index_label="#Names")




# =====
# Formatting
# =====
# Creates a unique identifier
def get_unique_identifier():
    return uuid.uuid4().hex
# Infer compression
def infer_compression(path:str):
    path = format_path(path)
    compression = None
    ext_zip = (".zip")
    ext_gzip = (".gz", ".gzip", ".pgz")
    ext_bz2 = (".bz2", ".bzip2", ".pbz2")
    if path.endswith(ext_gzip):
        compression= "gzip"
    if path.endswith(ext_bz2):
        compression = "bz2"
    if path.endswith(ext_zip):
        compression = "zip"
    return compression
# Format filename
def format_filename(name, replace="_"):
    Ar_name = np.array(list(name))
    idx_nonalnum = [*map(lambda x: x.isalnum() == False,  Ar_name)]
    Ar_name[idx_nonalnum] = replace
    return "".join(Ar_name)


# Get repr for custom classes
def get_repr(class_name, instance_name=None, *args):
    header = "{}(name = {})".format(class_name, instance_name)
    info = format_header(header)
    for field in args:
        info += "\n\t* {}".format(field)
    return info


# ======
# Matplotlib and colors
# ======
# Format handles for a matplotlib legend
def format_mpl_legend_handles(cdict_handles, label_specific_kws=None, marker="s", markeredgecolor="black", markeredgewidth=1,**args ):
    """
    More info on parameters: https://matplotlib.org/api/_as_gen/matplotlib.lines.Line2D.html#matplotlib.lines.Line2D.set_marker
    Usage: plt.legend(*format_mpl_legend_handles(cdict_handles),
                      loc="lower center",
                      bbox_to_anchor=(0.5,-0.15),
                      fancybox=True, shadow=True,
                      prop={'size':15})

    Input: cdict_handles = Dictionary object of {label:color}
    Output: Tuple of handles and labels
    """
    cdict_handles = OrderedDict(cdict_handles)

    handle_kws = {"marker":marker, "markeredgecolor":markeredgecolor, "markeredgewidth":markeredgewidth, "linewidth":0}
    handle_kws.update(args)

    labels = list(cdict_handles.keys())
    if label_specific_kws is not None:
        label_specific_kws = dict(label_specific_kws)
        assert set(labels) == set(label_specific_kws.keys()), f"If `label_specific_kws` is not None then it must have all elements from `cdict_handles`"
    else:
        label_specific_kws = {label:handle_kws for label in labels}

    handles = list()
    for label, color in cdict_handles.items():
        handle = plt.Line2D([0,0],[0,0], color=color, **label_specific_kws[label])
        handles.append(handle)
    return (handles, labels)

# Convert RGB ==> RGBA
def rgb_to_rgba(u, alpha=1.0, missing_alpha=1.0, name=None, verbose=True):
    """
    `u` is a pd.Series of {label:(r,g,b)}
    `alpha` can be a pd.Series, dict, or scalar
    """
    # Convert to Series
    u = pd.Series(u)
    assert len(u.values[0]) == 3, "Not all values are (r,g,b)"
    if is_dict(alpha):
        alpha = pd.Series(alpha, name="alpha")

    if is_number(alpha):
        assert 0.0 <= alpha <= 1.0, "`alpha` should be between 0 and 1"
        alpha = pd.Series([float(alpha)]*u.size, index=u.index, name="alphas")
    # Missing alpha values
    alpha = alpha[u.index]
    if pd.isnull(alpha).sum() > 0:
        alpha[alpha.isnull()] = missing_alpha
        if verbose:
            print(f"Warning: Filling in missing alpha values as {missing_alpha}", file=sys.stderr)
    # Convert to RGBA
    return pd.Series(map(lambda x: (*u[x],alpha[x]), u.index), index=u.index, name=name)

# Map Colors
def map_colors(y:pd.Series,
               mode=0,
               alpha=1.0,
               format="rgba",
               outlier_value=-1,
               outlier_color="white",
               palette="hls",
               cmap=None,
               desat=None,
               vmin=None,
               vmax=None,
               missing_alpha=1.0,
               name=None,
               ):
    """
    type(u) should be pd.Series for mode=1 but mode=0 can be any iterable

    mode=0: maps `u` to `n` colors where `n` is number of elements [e.g. a list of `n` clusters]
    mode=1: maps `u` to `m` colors where `m` is number of unique elements [e.g. a target vector]
    mode=2: maps `u` dict of groups -> values [e.g. a dictionary of clusters]
    mode=3: maps `u` dict of values to continuous colormap [e.g. a dictionary]
    mode=4: returns tuple of cmap and lmap
    [hex, rgb, rgba]

    `cmap` only works with mode == 3
    """
    # Nested kws
    nested_kws = dict(outlier_value=outlier_value, outlier_color=outlier_color, palette=palette, desat=desat)

    # Convert to pd.Series
    y = pd.Series(y)

    # Expand {group:iterable} to series
    if mode == 2:
        y = dict_expand(y.to_dict(), into=pd.Series)

    # One-to-one color mapping
    if mode==0:
        d_class_color = OrderedDict(zip(sorted(y.tolist()), sns.color_palette(palette=palette, n_colors=y.size, desat=desat)))
        if outlier_value in d_class_color:
            d_class_color[outlier_value] = outlier_color

        # Convert to series and format output
        class_colors = pd.Series(d_class_color, name=name).sort_index()
        if format == "hex":
            return class_colors.map(rgb2hex)
        if format == "rgb":
            return class_colors
        if format == "rgba":
            return rgb_to_rgba(class_colors, alpha=alpha, missing_alpha=missing_alpha, verbose=False)

    # Vertical list color mapping
    if mode in {1,2}:
        class_colors = map_colors(y.unique(),  mode=0, alpha=alpha, format="rgb", **nested_kws)
        obsv_colors = y.map(lambda x: class_colors[x])
        obsv_colors.name = name
        if format == "rgb":
            return obsv_colors
        if format == "rgba":
            return rgb_to_rgba(obsv_colors, alpha=alpha, missing_alpha=missing_alpha, name=name, verbose=False)
        if format == "hex":
            return obsv_colors.map(rgb2hex)

    # Continuous data
    if mode==3:
        # Converging or diverging color palettes
        return scalarmapping_from_data(y, cmap=cmap, vmin=vmin, vmax=vmax, alpha=alpha, format=format)

    # Return color mapping for vertical list and legend
    if mode == 4:
        obsv_colors = map_colors(y, mode=1, alpha=1.0, format=format, **nested_kws)
        class_colors = map_colors(y.unique(), mode=0, alpha=alpha, format=format, **nested_kws)
        return (obsv_colors, class_colors)

# Determine continuous data type
def infer_continuous_type(data:pd.Series):
    assert np.all(data.map(is_number)), "All values in `data` must be numerical"

    _vmin = data.min()
    _vmax = data.max()

    # Conditions
    conditions = defaultdict(list)
    conditions["A"].append(_vmin < 0)
    conditions["A"].append(_vmax > 0)

    # +/-
    if all(conditions["A"]):
        return "diverging"
    # Other
    else:
        return "sequential"

# Determining vmin and vmax
def infer_vmin_vmax(data:pd.Series, continuous_type="infer"):
    vmin = None
    vmax = None
    # Infer continuous type
    if continuous_type in ["infer", None]:
        continuous_type = infer_continuous_type(data)
    # +/-
    if continuous_type == "diverging":
        vmax = data.abs().max()
        vmin = -vmax
    # Other
    if continuous_type == "sequential":
        vmax = data.max()
        vmin = data.min()
    assert all(map(bool, [vmin,vmax])), "`vmin` and `vmax` should not be None at this point.  Please check `infer_continuous_type`"
    return vmin, vmax

# Determing converging or diverging colormap
def infer_cmap(data:pd.Series, diverging=TealRose_3.mpl_colormap, sequential=Tempo_8.mpl_colormap):
    """
    diverging and sequential are both mpl colormaps or compatible
    """
    continuous_type = infer_continuous_type(data)
    return {"diverging":diverging, "sequential":sequential}[continuous_type]

# Mapping scalar values to colors
def scalarmapping_from_data(data:pd.Series, cmap=None, vmin=None, vmax=None, alpha=1.0, mode=1, format="rgba", name=None, missing_alpha=1.0):
    """
    Function that takes in numerical values and converts them to a pd.Series color mapping
    Must provide either `data` or (`cmap`, `vmin`, and `vmax`).  The latter is only compatible with mode == 0
    Input:
        type(data) is pd.Series
        cmap = mpl or sns cmap object
        vmin = minimumum value
        vmax = maximumum value
    mode = 0:
        Output: Fitted scalarmappable
    mode = 1:
        Output: Transformed values to colors
    Format:
        `rgba`, `rgb`, or `hex`
    Log:
        09.28.2017
    """
    # Useable data
    data = pd.Series(data).dropna()

    # Condtions
    conditions = defaultdict(list)

    # vmin/vmax
    continuous_type = infer_continuous_type(data)
    conditions["A"].append(vmin in ["infer", None])
    conditions["A"].append(vmax in ["infer", None])
    if any(conditions["A"]):
        assert all(conditions["A"]), "If `vmin` or `vmax` is `infer` then both must be `infer`.  Explicitly state  `vmin` and `vmax`"
        vmin, vmax = infer_vmin_vmax(data, continuous_type=continuous_type)

    # Colormap
    if cmap is None:
        cmap = infer_cmap(data)

    # Scalar Mapper
    sm = plt.cm.ScalarMappable(cmap=cmap, norm=plt.Normalize(vmin=vmin, vmax=vmax))

    # Outputs
    if mode == 0:
        return sm
    if mode == 1:
        obsv_colors = data.map(lambda x: sm.to_rgba(x)[:-1])
        if format == "rgb":
            return obsv_colors
        if format == "rgba":
            return rgb_to_rgba(obsv_colors, alpha=alpha, missing_alpha=missing_alpha, name=name, verbose=False)
        if format == "hex":
            return obsv_colors.map(rgb2hex)

# Add colorbar from the data
def add_cbar_from_data(fig, cmap, data=None, vmin=None, vmax=None, cbar_pos=[0.925, 0.1, 0.015, 0.8], label="infer", cbar_kws=dict(), tick_kws=dict(), **args):
    """
    Input:
        `fig` = a matplotlib figure
        `cmap` = matplotlib or seaborn cmap
        `data` = pd.Series of data
    Output:
        cbar, ax_colorbar
    Log:
        03.11.2019 Changed `u` to `data` and adjusted the style for conditions, now returns cbar, ax_colorbar
        09.28.2017 removed inactivate arguments, made fig, cmap, and u mandatory, added cbar_pos for customization in cbar position
    """
    conditions = defaultdict(list)
    for condition in ["A","B"]:
        conditions[condition].append(fig is not None)
        conditions[condition].append(cmap is not None)
    conditions["A"].append(data is not None)
    conditions["B"].append(vmin is not None)
    conditions["B"].append(vmax is not None)

    assert all(conditions["A"]) or all(conditions["B"]), "Must input either `fig`, `cmap`, and `data` arguments or `fig`, `cmap`, `vmin` and `vmax`"

    # ax_colorbar.tick_params
    _tick_kws = {}# {"size":15}
    _tick_kws.update(tick_kws)
    # cbar.set_label
    _cbar_kws = {"fontsize":18, "rotation":270, "labelpad":25}
    _cbar_kws.update(cbar_kws)
    # From data
    if all(conditions["A"]):
        if label == "infer":
            label = data.name

        sm = scalarmapping_from_data(data, cmap=cmap, vmin=vmin, vmax=vmax, mode=0)
    else:
        sm = plt.cm.ScalarMappable(cmap=cmap, norm=plt.Normalize(vmin=vmin, vmax=vmax))

    # Set parameters
    ax_colorbar = fig.add_axes(cbar_pos)
    sm._A = []

    # Color bar
    cbar = fig.colorbar(sm, cax=ax_colorbar, **args)
    ax_colorbar.tick_params(**_tick_kws)
    # Labels
    if label:
        cbar.set_label(label, **_cbar_kws)
    return cbar, ax_colorbar

# Configure Scatter Plot Data
def configure_scatter(

                     # Main
                     data,
                     x=None,
                     y=None,
                     z=None,
                     c="gray",
                     s=None,
                     alpha=1.0,

                     # Aspects
                     cmap = None,
                     vmin=None,
                     vmax=None,
                     dimensions=2,

                     # Missing data
                     missing_color='#FFFFF0',
                     missing_alpha=1.0,
                     verbose=True,
                     ):

    """
    Preprocessing for 2-/3D scatter plot data.  Formats color, size, and alpha values.  Designed for matplotlib input but should be able to be universal for Plot.ly and Bokeh
    """
     # ==================================================
    # DataFrame
    # ==================================================
    # Default x and y axis
    if x is None:
        x = data.iloc[:,0]
    if y is None:
        y = data.iloc[:,1]
    if type(x) == str:
        x = data[x]
    if type(y) == str:
        y = data[y]
    if dimensions == 3:
        if z is None:
            z = data.iloc[:,2]
        if type(z) == str:
            z = data[z]
    # Coordinates
    if dimensions == 2:
        coords = pd.DataFrame([x,y], index=["x","y"]).T
    if dimensions == 3:
        coords = pd.DataFrame([x,y,z], index=["x","y","z"]).T

    # Defaults
    if c is None:
        c = "gray"
    if s is None:
        N = data.shape[0]
        s = np.square((1/np.log(N))*50)

    # Is color a labeled column?
    if type(c) == str:
        # Use values from color column
        if c in data.columns:
            c = data[c]
        # Color all values same color
        else:
            c = pd.Series([c]*len(x), index=data.index)

    # Convert colors to pd.Series
    c = pd.Series(c)[data.index]
    scalar_vector = None

    # Continuous mapping for colors
    if is_number(c.values[0]):
        if pd.isnull(c.values[0]) == False:
            scalar_vector = c.copy()
            if cmap is None:
                cmap = infer_cmap(c)
                c = scalarmapping_from_data(c, cmap=cmap, vmin=vmin, vmax=vmax, alpha=1.0, format="rgb", mode=1, name=c.name)

    # Fill missing colors
    if pd.isnull(c).sum() > 0:
        c[c.isnull()] = missing_color
        if verbose:
            print(f"Warning: Filling in missing color values as {missing_color}", file=sys.stderr)

    # Set up alpha values
    # If Hex string, convert to RGB
    if isinstance(c.values[0], (str, bytes)):
        c = c.map(to_rgb)

    # If RGB or RGBA
    # WARNING: Cannot handle an instance where there is a subset of the indices and the values of the subset are floats/ints
    if alpha is not None:
        if hasattr(c.values[0], '__iter__'):
            # If RGB then add alpha to RGBA
            if len(c.values[0]) == 3:
                if is_number(alpha):
                    assert 0.0 <= alpha <= 1.0, "`alpha` should be between 0 and 1"
                    alpha = pd.Series([float(alpha)]*c.size, index=c.index, name="alphas")
                # Fill missing alpha values
                if pd.isnull(alpha).sum() > 0:
                    alpha[alpha.isnull()] = missing_alpha
                    if verbose:
                        print(f"Warning: Filling in missing alpha values as {missing_alpha}", file=sys.stderr)
                # Convert to RGBA
                c = pd.Series(map(lambda x: (*c[x],alpha[x]), c.index), index=c.index, name=c.name)
    # Size
    if s is not None:
        if type(s) == str:
            if s in data.columns:
                s = data[s]
        if is_number(s):
            s = pd.Series([s]*len(x), index=data.index, name="sizes")
        # Convert sizes to pd.Series
        s = pd.Series(s, name="sizes")[data.index]
        # Fill in missing sizes
        if pd.isnull(s).sum() > 0:
            med = s.dropna().median()
            s[s.isnull()] = med
            if verbose:
                print(f"Warning: Filling in missing size values as {med}", file=sys.stderr)
    return {
            "coords":coords,
            "colors":c,
            "sizes":s,
            "cmap":cmap,
            "scalar_vector":scalar_vector,
           }

# # Colors class object
# class Hue(str):
#     # Built in
#     def __new__(cls, color, name=None, **metadata):
#         assert is_color(color), "`{}` is not a color".format(color)
#         cls.color = rgb2hex(to_rgb(color))
#         cls.name = name
#         cls.metadata = metadata
#         return super(Hue, cls).__new__(cls, cls.color)
#     def __repr__(self):
#         return self.color
#     def __str__(self):
#         return self.color
#     # View
#     def view(self):
#         sns.palplot([self.color])
#     # Conversion
#     def to_hex(cls):
#         return cls.color
#     def to_rgb(cls, into=tuple):
#         return into(to_rgb(cls.color))
#     def to_rgba(cls, alpha=None, into=tuple):
#         return into(to_rgba(cls.color, alpha=alpha))
#     def to_rgb_255(cls, into=tuple):
#         return into(np.asarray(cls.to_rgb())*255)
#     def to_rgba_255(cls, alpha=None, into=tuple):
#         return into(np.asarray(cls.to_rgba(alpha=alpha))*255)

# Color collections
class Chromatic(object):
    """
    # =================
    # Continuous colors
    # =================
    colors_sepal_length = Chromatic.from_continuous(y=X_iris["sepal_length"], name="sepal_length", cmap=plt.cm.magma)
    print(colors_sepal_length.obsv_colors.iloc[:3])
    # iris_0    #deb5e8
    # iris_1    #efc5e6
    # iris_2    #f9d7e7
    # Name: sepal_length, dtype: object
    colors_sepal_length.__dict__.keys()
    # dict_keys(['name', '__synthesized_', 'vmin', 'vmax', 'cmap', 'continuous_type', 'format', 'mode', 'obsv_colors', 'y'])

    # ==============
    # Classes colors
    # ==============
    colors_species = Chromatic.from_classes(y_iris, name="iris species", palette="Set2")
    print(colors_species.class_colors)
    # setosa        #66c2a5
    # versicolor    #fc8d62
    # virginica     #8da0cb
    print(colors_species.obsv_colors.iloc[:3])
    # iris_0    #66c2a5
    # iris_1    #66c2a5
    # iris_2    #66c2a5
    # Name: Species, dtype: object
    colors_species.__dict__.keys()
    # dict_keys(['name', '__synthesized_', 'format', 'mode', 'obsv_colors', 'class_colors', 'y', 'palette', 'color_kws'])
    """
    def __init__(self, **kwargs):
        attributes = {"name":None, "__synthesized_":datetime.datetime.utcnow()}
        attributes.update(kwargs)
        for k,v in attributes.items():
            setattr(self, k,v)

    @classmethod
    def from_classes(cls, y:pd.Series, name=None, palette="hls", class_type=None, obsv_type=None, format="hex", outlier_value=-1, outlier_color='white', color_kws=dict()):
        if is_dict(y):
            y = pd.Series(y)
        obsv_colors, class_colors = map_colors(y, mode=4, format=format, palette=palette, outlier_value=outlier_value, outlier_color=outlier_color, **color_kws)
        attributes = { "name":name, "format":format, "mode":"classes", "obsv_colors":obsv_colors,"class_colors":class_colors, "y":y.copy(), "palette":palette,"color_kws":color_kws}
        return cls(**attributes)

    @classmethod
    def from_continuous(cls, y:pd.Series, name=None, cmap="infer", vmin="infer", vmax="infer", format="hex", **attrs):
        if is_dict(y):
            y = pd.Series(y)
        assert np.all(y.map(is_number)), "All values in `y` must be numerical"

        conditions = defaultdict(list)

        # vmin/vmax
        continuous_type = infer_continuous_type(y)
        conditions["A"].append(vmin in ["infer", None])
        conditions["A"].append(vmax in ["infer", None])
        if any(conditions["A"]):
            assert all(conditions["A"]), "If `vmin` or `vmax` is `infer` then both must be `infer`.  Explicitly state  `vmin` and `vmax`"
            vmin, vmax = infer_vmin_vmax(y, continuous_type=continuous_type)

        # cmap
        if cmap in ["infer", None]:
            cmap = infer_cmap(y)

        # Get colors
        obsv_colors = map_colors(y, mode=3, vmin=vmin, vmax=vmax, format=format, cmap=cmap)
        attributes = {"vmin":vmin, "vmax":vmax, "cmap":cmap, "continuous_type":continuous_type, "name":name, "format":format, "mode":"continuous", "obsv_colors":obsv_colors, "y":y.copy()}
        return cls(**attributes)

    def plot_colors(self, color_type="infer", size=1):
        accepted_colortypes = {"infer",None,"classes", "obsvs", "leaves", "obsvervations"}
        assert color_type in accepted_colortypes, f"`color_type` must be one of the following: {accepted_colortypes}"
        if color_type in ["infer",None]:
            if hasattr(self, "class_colors"):
                color_type = "classes"
            else:
                color_type = "obsvs"

        if color_type in ["obsvs", "observations", "leaves"]:
            if hasattr(self, "continuous_type"):

                colors = self.obsv_colors[self.y.sort_values().index]
            else:
                colors = self.obsv_colors.sort_values()
        if color_type == "classes":
            assert hasattr(self, "class_colors"), "There isn't a `class_colors` object.  Is this continuous data?  Use `color_type = 'obsvs' instead."
            colors = self.class_colors
        sns.palplot(colors, size=size)

# ===========
# Coordinates
# ===========
# Get 3d contour interpolation
def get_coords_contour(x,y,z,resolution = 50, method_iterpolation='linear'):
    # Adapted from @user10140465
    # https://stackoverflow.com/questions/52935143/how-to-do-a-contour-plot-from-x-y-z-coordinates-in-matplotlib-plt-contourf-or/52937196#52937196
    resolution = f"{resolution}j"
    X,Y = np.mgrid[min(x):max(x):complex(resolution),
                   min(y):max(y):complex(resolution)]
    points = list(zip(x,y))
    Z = griddata(points, z, (X, Y), method=method_iterpolation)
    return X,Y,Z

# Format centroids for group of indices
def get_coords_centroid(df_xy, groups, metric=np.mean, into=list, dimensions=2):
    """
    groups: {group_A: [obsv_1, obsv_2], group_B: [obsv_3, obsv_4]}
    Output:
    [(group_i, (x_i,y_i))]

    Note: This only works for 2D
    Previous names: centroids_2d
    """
    assert dimensions == 2, "Currently only is available for 2 dimensional datasets for plotting purposes"

    # Compute centroid for 2 vectors based on groups
    def _compute(df_xy, groups, metric):
        if is_nonstring_iterable(groups):
            groups = dict(groups)
        groups = {r:p for p,q in groups.items() for r in q}
        centroids_xy = df_xy.groupby(groups).apply(metric, axis=0)
        if isinstance(centroids_xy, pd.Series):
            centroids_xy = pd.DataFrame.from_items(zip(centroids_xy.index, centroids_xy.values)).T
            centroids_xy.columns = ["x","y"]
        return centroids_xy
    df_centroid_xy = _compute(df_xy=df_xy, groups=groups, metric=metric)

    if into == pd.DataFrame:
        return df_centroid_xy
    else:
        return into([*zip(df_centroid_xy.index, zip(df_centroid_xy.iloc[:,0], df_centroid_xy.iloc[:,1]))])

# Format ellipse attributes
def get_parameters_ellipse(df_xy, groups, metric=np.mean, n_std=3, into=pd.DataFrame):
    """
    groups: {group_A: [obsv_1, obsv_2], group_B: [obsv_3, obsv_4]}

    Previous names: centroids_ellipse
    """
    # Compute ellipse attributes
    def _compute(df_xy, idx, metric, n_std):
        def eigsorted(cov):
            vals, vecs = np.linalg.eigh(cov)
            order = vals.argsort()[::-1]
            return vals[order], vecs[:,order]
        # Data
        x =  df_xy.iloc[:,0][idx].as_matrix()
        y =  df_xy.iloc[:,1][idx].as_matrix()
        # Reduce
        x_reduce = metric(x)
        y_reduce = metric(y)
        # Diff
        cov = np.cov(x, y)
        vals, vecs = eigsorted(cov)
        theta = np.degrees(np.arctan2(*vecs[:,0][::-1]))
        width, height = 2 * n_std * np.sqrt(vals)

        return OrderedDict([
                          ("x", x_reduce),
                          ("y", y_reduce),
                          ("theta", theta),
                          ("width", width),
                          ("height", height),
                            ])
    ellipse_data = list()
    for group, idx in groups.items():
        ellipse_data.append(pd.Series(_compute(df_xy=df_xy, idx=idx, metric=metric, n_std=n_std), name=group))
    return into(ellipse_data)


# ===========
# Logging
# ===========
# Log files
def create_logfile(name, path,level=logging.INFO, mode="w"):
    """Function setup as many loggers as you want"""
    # https://stackoverflow.com/questions/11232230/logging-to-two-files-with-different-settings

    formatter = logging.Formatter(fmt='%(asctime)s %(levelname)s %(message)s', datefmt='%H:%M:%S')

    handler = logging.FileHandler(path, mode=mode)
    handler.setFormatter(formatter)

    logger = logging.getLogger(name)
    logger.setLevel(level)
    logger.addHandler(handler)

    return logger

# Determine whether or not logfile should be overwritten or appended
def determine_mode_for_logfiles(path, force_overwrite):
    if force_overwrite:
        mode_for_logfiles = "w"
    else:
        if os.path.exists(path):
            mode_for_logfiles = "a"
        else:
            mode_for_logfiles = "w"
    return mode_for_logfiles


# ===========
# Pandas
# ===========
# Prepend level to pandas index
def pd_prepend_level_to_index(x, level_value, level_name=None):
    """
    Prepend a level to a pd.Series or pd.DataFrame
    
    Usage: 
    data = list("ATG")
    x = pd.Series(
        data, 
        index=pd.Index(range(0,len(data)), name="Position"),
        name="START",
    )
    print(x)
    # Position
    # 0    A
    # 1    T
    # 2    G
    # Name: START, dtype: object
    print(pd_prepend_level_to_index(x, "Chr1", level_name="Chromosome"))
    # Chromosome  Position
    # Chr1        0           A
    #             1           T
    #             2           G
    # Name: START, dtype: object
    
    Adapted from the following source:
    https://stackoverflow.com/questions/14744068/prepend-a-level-to-a-pandas-multiindex
    """
    return pd.concat([x], keys=[level_value], names=[level_name])

# Collapse a pandas series
def pd_series_collapse(u:pd.Series, into=pd.Series, type_collection=set):
    """
    Input: Pandas Series object with redundant values (e.g. iris targets)
    Output: Object organized by group

    e.g., a:b to {b:[x,y,z]} or another pd.Series with a similar vibe
    """
    data = into(dict(map(lambda x: (x[0], type_collection(x[1].index)), u.groupby(u))))
    if hasattr(data, "name"):
        data.name = u.name
    return data

# Convert pandas series to a groupby object and then a dataframe of the groups
def pd_series_to_groupby_to_dataframe(data, by, fillna=None):
    df = pd.DataFrame([*map(lambda group: pd.Series(group[1], name=group[0]).reset_index(drop=True), data.groupby(by))]).T
    if fillna is not None:
        df = df.fillna(fillna)
    return df

# Filter pandas series
def pd_series_filter(func, u):
    mask = u.map(func)
    return u[mask]

# Query a pandas dataframe
def pd_dataframe_query(df:pd.DataFrame, conditions, mode="all"):
    """
    conditions = dict with key as column and value as the condition function

    conditions = {"run": lambda id_run: id_run in [3,4,5,9], "sample-category": lambda x: x in ["unknown", "control"]}
    pd_dataframe_query(df_meta, conditions)
    """
    number_of_conditions = len(conditions)
    assert number_of_conditions > 0, "Must provide at least one conditional"
    assert mode in {"any", "all"}, "`mode` must be either 'any' or 'all'"
    if mode == "all":
        tol = number_of_conditions
    if mode == "any":
        tol = 1

    masks = list()
    for key, condition in conditions.items():
        masks.append(df[key].map(condition))
    idx_query = pd.DataFrame(masks).sum(axis=0)[lambda x: x >= tol].index
    return df.loc[idx_query,:]

# Matrix Multiplication
def pd_dataframe_matmul(A:pd.DataFrame,B:pd.DataFrame):
    """
    A.shape (n,m)
    B.shape (m,p)
    Output: (n,p)
    """
    return pd.DataFrame(np.matmul(A, B.loc[A.columns,:]), index=A.index, columns=B.columns)

# Extend a DataFrame
def pd_dataframe_extend_index(index_extended, df=None, fill=np.nan, axis=0):
    """
    Extend the index of pd.DataFrame like pandas used to do before 23.4
    """
    if df is None:
        df = pd.DataFrame()
    if axis == 0:
        idx_extend = set(index_extended) - set(df.index)
        A = np.empty((len(idx_extend), df.shape[1]))
        A[:] = np.nan
        return pd.concat([df, pd.DataFrame(A, index=idx_extend, columns=df.columns)]).fillna(fill)
    if axis == 1:
        idx_extend = set(index_extended) - set(df.columns)
        A = np.empty(( df.shape[0], len(idx_extend)))
        A[:] = np.nan
        return pd.concat([df, pd.DataFrame(A, index=df.index, columns=idx_extend)]).fillna(fill)

# Drop duplicates index
def pd_dropduplicates_index(data, keep="first", axis=0):
    if axis in {0, None}:
        return data[~data.index.duplicated(keep=keep)]
    if axis == 1:
        data = data.T
        return data[~data.index.duplicated(keep=keep)].T

# # =======
# # Filters
# # =======
# Filter composition 2D data
def filter_compositional(
    X:pd.DataFrame,
    tol_depth=None,
    tol_prevalence=None,
    tol_richness=None,
    tol_count=None,
    order_of_operations=["depth", "prevalence", "richness","count"],
    mode="highpass",
    interval_type="closed",
    ):
    """
    Input: pd.DataFrame (rows=obsvs, cols=attrs)
    Output: pd.DataFrame filtered
    depth:  The minimum depth (sum per row) (axis=0)
    prevalence: The minimum number of observations that must contain the attribute (axis=1)
    count: The minimum number of total counts contained by the attribute (axis=1)
    richess: The minimum number of detected attributes (axis=0)
    """
    assert_acceptable_arguments(query=order_of_operations,target=["depth", "prevalence", "richness","count"], operation="le")
    assert_acceptable_arguments(query=[mode],target=["highpass", "lowpass"], operation="le")
    assert_acceptable_arguments(query=[interval_type],target=["closed", "open"], operation="le")


    def _get_elements(data,tol,operation):
        return data[lambda x: operation(x,tol)].index

    def _filter_depth(X, tol, operation):
        data = X.sum(axis=1)
        return X.loc[_get_elements(data, tol, operation),:]

    def _filter_prevalence(X, tol, operation):
        conditions = [
            isinstance(tol, float),
            0.0 < tol <= 1.0,
        ]

        if all(conditions):
            tol = round(X.shape[0]*tol)
        data = (X > 0).sum(axis=0)
        assert tol <= X.shape[0], "If prevalence is an integer ({}), it cannot be larger than the number of samples ({}) in the index".format(tol, X.shape[0])
        return X.loc[:,_get_elements(data, tol, operation)]

    def _filter_richness(X, tol, operation):
        data = (X > 0).sum(axis=1)
        return X.loc[_get_elements(data, tol, operation),:]

    def _filter_count(X, tol, operation):
        data = X.sum(axis=0)
        return X.loc[:,_get_elements(data, tol, operation)]
    if interval_type == "closed":
        operations = {"highpass":operator.ge, "lowpass":operator.le}
    if interval_type == "open":
        operations = {"highpass":operator.gt, "lowpass":operator.lt}

    # Defaults
    if tol_richness is None:
        tol_richness = 1
    if tol_count is None:
        tol_count = 1
    if tol_prevalence is None:
        tol_prevalence = 1

    functions = dict(zip(["depth", "prevalence", "richness","count"], [_filter_depth, _filter_prevalence, _filter_richness, _filter_count]))
    thresholds = dict(zip(["depth", "prevalence", "richness","count"], [tol_depth, tol_prevalence, tol_richness, tol_count]))


    for strategy in order_of_operations:
        tol = thresholds[strategy]
        if tol is not None:
            X = functions[strategy](X=X,tol=tol, operation=operations[mode])

    return X


# ==============================================================================
# Imports
# ==============================================================================
# Built-ins
import os, sys, warnings, re, time, copy
from collections import defaultdict, OrderedDict
import pandas as pd
import numpy as np
import networkx as nx
import matplotlib.pyplot as plt

# Stats
import statsmodels.api as sm
import statsmodels.formula.api as smf
from statsmodels.sandbox.stats.multicomp import multipletests
from sklearn.metrics import r2_score
from scipy.stats import linregress

# Parallel
from joblib import Parallel, delayed

# Utility
from tqdm import tqdm

# Soothsayer
from ..utils import *
from ..io import write_object
from ..statistics import p_adjust
from ..visuals import draw_networkx_labels_with_box, plot_scatter

__all__ = ["MixedEffectsLinearModel", "GeneralizedLinearModel"]
__all__ = sorted(__all__)

# Core linear regression methods
class CoreLinearModel(object):

    # Check the regression variable format
    def check_regression_variable_format(self, label:str):
        alphanumeric = all(map(lambda x:x.isalnum() or (x == "_"), label))
        starts_with_alpha = label[0].isalpha()
        return all([alphanumeric, starts_with_alpha])

    # Relabel variables and attributes
    def relabel_attributes(self, df:pd.DataFrame, attribute_prefix="attr_"):
        n,m = df.shape
        df_relabeled = df.copy()
        idx_attrs_relabeled = [*map(lambda j:"{}{}".format(attribute_prefix, j), range(df.shape[1]))]
        encoding = dict(zip(df.columns, idx_attrs_relabeled))
        decoding = dict(zip(idx_attrs_relabeled, df.columns))
        df_relabeled.columns = idx_attrs_relabeled
        return df_relabeled, encoding, decoding

    # Decode the relabeled index
    def decode_index(self, indice):
        level_0, level_1 = indice
        id_attribute = re.split('\[T\..+]', level_0)[0]
        if id_attribute in self.decoding_:
            level_0 = level_0.replace(id_attribute, self.decoding_[id_attribute])
        return (level_0, level_1)

    # Convert a statsmodels table to pd.Series
    def statsmodels_table_to_pandas(self, model_results, name, multiple_comparison_method="fdr", residual_type="infer"):
        if is_query_class(model_results, "mixed"):
            residual_type = "resid"
        else:
            residual_type = "resid_response"
        assert hasattr(model_results, residual_type)
        statistics_data = OrderedDict([
            ("t_values",model_results.tvalues),
            ("p_values",model_results.pvalues),
            ("std_err", model_results.bse),
            ("coef", model_results.params),
            ("[0.025", model_results.conf_int().iloc[:,0]),
            ("0.975]", model_results.conf_int().iloc[:,1]),
            ])
        if multiple_comparison_method is not None:
            statistics_data["{}_values".format(multiple_comparison_method)] = p_adjust(model_results.pvalues, method=multiple_comparison_method)




        # Model
        model_data = {
            "scale":model_results.scale,
            "likelihood":model_results.llf,
            "converged":model_results.converged,
            "method":model_results.method,
            "num_observations":model_results.nobs,
            "r_squared":r2_score(model_results.y_actual, model_results.fittedvalues),
        }
        residuals = getattr(model_results, residual_type)
        model_data["mse"] = np.power(residuals, 2).mean()

        Se_statistics = pd.DataFrame(statistics_data).stack()
        Se_model = pd.Series(model_data)
        Se_model.index = pd.MultiIndex.from_tuples(Se_model.index.map(lambda x: ("Model", x)), names=[None,None])
        synopsis = pd.concat([Se_statistics, Se_model])
        synopsis.name = name
        return synopsis

    # Predict for a particular attribute
    def predict(self, attribute:str, Y:pd.DataFrame):
        assert self.fitted, "Please fit model first."
        Y = Y.copy()
        assert attribute in self.model_results_, "`{}` not in training data".format(attribute)
        # assert set(Y.columns) <= set(self.encoding_.keys()), "Not all variables in Y were used during fitting"
        Y.columns = Y.columns.map(lambda x: self.encoding_[x] if x in self.encoding_ else x)
        model = self.model_results_[attribute]
        return  model.predict(exog=Y)

    # Create directed networkx graph
    def create_graph(self, graph=None, test_field="fdr_values", tol_test=0.05, exclude_attributes:list=None, **attrs):
        assert self.fitted, "Please fit model before continuing."
        if tol_test is None:
            tol_test = 1.0
        assert 0 < tol_test <= 1.0, "0 < `tol_test` ≤ 1.0 "
        attrs.update({"test_field":test_field, "tol_test":tol_test})

        # Construct graph
        if graph is None:
            graph = nx.OrderedDiGraph(name=self.name, **attrs)
        assert isinstance(graph, (nx.DiGraph, nx.OrderedDiGraph)), "`graph` must either be DiGraph or OrderedDiGraph"
        self.graph = graph

        # Drop fields and attributes
        index = self.synopsis_.columns.get_level_values(0)
        drop_fields = list()
        for id_field in ["Intercept", "Group Var"]:
            if id_field in index:
                drop_fields.append(id_field)

        drop_attributes = list()
        if exclude_attributes is not None:
            drop_attributes += list(exclude_attributes)

        # Synopsis
        df_synopsis = self.synopsis_.copy().drop(drop_attributes, axis=0).drop(drop_fields, axis=1)
        assert test_field in df_synopsis.columns.get_level_values(1), "`{}` not available in `synopsis_`"

        # Add edges to graph
        for id_attribute, data in df_synopsis.iterrows():
            test_values = data[data.index.map(lambda x:x[1] == test_field)]
            mask_tol = test_values < tol_test
            if np.any(mask_tol):
                signficant_variables = test_values[mask_tol].index.get_level_values(0)
                coefficients = data[[*map(lambda x: (x, "coef"), signficant_variables)]]
                for (id_variable, _), coef in coefficients.iteritems():
                    self.graph.add_edge(id_variable, id_attribute, weight=np.abs(coef), sign=np.sign(coef), **data[id_variable].to_dict())
        # Check nodes
        number_of_nodes = self.graph.number_of_nodes()
        number_of_edges = self.graph.number_of_edges()

        if self.verbose:
            if number_of_nodes == 0:
                warnings.warn("Number of nodes is 0 with {} < {}".format(test_field, tol_test))
            else:
                print("N(Nodes) = {}, N(Edges) = {}".format(number_of_nodes, number_of_edges), file=sys.stderr)

        return self.graph

    # Plot graphical visualization
    def plot_graph(self,
                   pos=None,
                   show_node_labels=False,
                   node_colors=None,
                   edge_colors=None,
                   variable_color="#808080",
                   color_negative="#278198",
                   color_positive="#dc3a23",
                   node_sizes=500,
                   title=None,
                   title_kws=dict(),
                   func_edgewidth=1.618,
                   node_alpha=0.618,
                   edge_alpha=0.618,
                   node_edgewidths=1,
                   node_edgecolors="white",
                   font_color="black",
                   node_box_facecolor="darkgray",
                   node_box_edgecolor="white",
                   legend=None,
                   legend_markerscale=2,
                   legend_kws=dict(),
                   node_kws=dict(),
                   node_label_kws=dict(),
                   edge_kws = dict(),
                   figsize=(13,8),
                  ):
        assert hasattr(self, "graph"), "Please create graph before continuing."
        attributes = set(self.synopsis_.index) & set(self.graph.nodes())
        variables = set(self.graph.nodes()) - attributes
        # =====
        # Nodes
        # =====
        # Colors
        if node_colors is None:
            assert is_color(variable_color), "`variable_color = {}` is not a color".format(variable_color)
            node_colors = "#000000"
        if is_color(node_colors):
            node_colors = pd.concat([
                pd.Series([node_colors]*len(attributes), index=attributes),
                pd.Series([variable_color]*len(variables), index=variables),
            ])
        node_colors = pd.Series(node_colors)
        assert set(node_colors.index) >= set(attributes), "Not all attributes have node colors"
        for id_variable in variables:
            if id_variable not in node_colors:
                node_colors[id_variable] = variable_color
        node_colors = node_colors[list(self.graph.nodes())]

        # Sizes
        if is_number(node_sizes):
            node_sizes = pd.concat([
                pd.Series([node_sizes]*len(attributes), index=attributes),
                pd.Series([0]*len(variables), index=variables),
            ])
        node_sizes = pd.Series(node_sizes)
        assert set(node_sizes.index) >= set(attributes), "Not all attributes have node sizes"
        for id_variable in self.fixed_effect_variables_:
            node_sizes[id_variable] = 0
        node_sizes = node_sizes[list(self.graph.nodes())]

        # Labels
        if show_node_labels:
            if show_node_labels == True:
                show_node_labels = list(attributes)
            show_node_labels = dict(zip(show_node_labels, show_node_labels))
        # =====
        # Edges
        # =====
        # Weights
        coefficients = pd.Series(OrderedDict([*map(lambda edge_data: ((edge_data[0], edge_data[1]), edge_data[2]["coef"]), self.graph.edges(data=True))]))
        weights = coefficients.abs()
        if func_edgewidth is not None:
            if is_number(func_edgewidth):
                constant = func_edgewidth
                func_edgewidth = lambda w: float(constant)
            weights = weights.map(func_edgewidth)

        # Colors
        if edge_colors is None:
            edge_colors = coefficients.copy()
            edge_colors[coefficients > 0] = color_positive
            edge_colors[coefficients < 0] = color_negative
            edge_colors[coefficients == 0] = "black"

        # ========
        # Plotting
        # ========
        _title_kws = {"fontsize":15, "fontweight":"bold"}
        _title_kws.update(title_kws)
        _legend_kws = {'fontsize': 15, 'frameon': True, 'facecolor': 'white', 'edgecolor': 'black', 'loc': 'center left', 'bbox_to_anchor': (1, 0.5)}
        _legend_kws.update(legend_kws)
        _node_kws = dict(node_color=node_colors.values, node_size=node_sizes.values, linewidths=node_edgewidths, edgecolors=node_edgecolors, alpha=node_alpha)
        _node_kws.update(node_kws)
        _node_label_kws = {"font_color":font_color, "font_size":12}
        _node_label_kws.update(node_label_kws)
        _edge_kws = dict(edge_color=edge_colors, width=weights, alpha=edge_alpha)
        _edge_kws.update(edge_kws)

        if pos is None:
            pos = nx.nx_agraph.graphviz_layout(self.graph, prog="neato")
        self.pos = pos
        with plt.style.context("seaborn-white"):
            fig, ax = plt.subplots(figsize=figsize)
            nx.draw_networkx_edges(self.graph, pos=pos, ax=ax, **_edge_kws)
            nx.draw_networkx_nodes(self.graph, pos=pos,  ax=ax, **_node_kws)
            if show_node_labels:
                nx.draw_networkx_labels(self.graph, pos=pos, labels=show_node_labels, ax=ax, **_node_label_kws)
            draw_networkx_labels_with_box(self.graph, pos=pos, labels = variables, ax=ax, font_color=font_color, node_box_facecolor=node_box_facecolor, node_box_edgecolor=node_box_edgecolor)
            # Legend
            if legend is not None:
                ax.legend(*format_mpl_legend_handles(pd.Series(legend)),  markerscale=legend_markerscale, **_legend_kws)
            if title is not None:
                ax.set_title(title, **_title_kws)
            ax.axis("off")
            return fig, ax

    # Plot residuals
    def plot_residuals(self,
                       attribute:str,
                       show_attribute=True,
                       show_reference_fit=False,
                        c = None,
                        alpha=0.8,
                       title = None,
                       legend = None,
                       title_kws = dict(),
                       legend_kws = dict(),
                       scatter_kws=dict(),
                       annot_kws = dict(),

                       **kwargs,
                      ):
        assert self.fitted, "Please fit model first."
        assert attribute in self.model_results_, "`{}` not in training data".format(attribute)




        # Defaults
        _annot_kws = dict(xytext=(12, -12), va='top', xycoords='axes fraction', textcoords='offset points',fontsize=15)
        _annot_kws.update(annot_kws)
        _scatter_kws = dict(alpha=alpha, c=c)
        _scatter_kws.update(scatter_kws)
        _title_kws = {"fontsize":15, "fontweight":"bold"}
        _title_kws.update(title_kws)
        _legend_kws = {'fontsize': 15, 'frameon': True, 'facecolor': 'white', 'edgecolor': 'black', 'loc': 'center left', 'bbox_to_anchor': (1, 0.5)}
        _legend_kws.update(legend_kws)

        # Data
        data = pd.concat([self.model_results_[attribute].fittedvalues.to_frame("ŷ"), self.X_[attribute].to_frame("y")], axis=1)
        data["residuals"] = self.residuals_[attribute]
        r2 = self.synopsis_.loc[attribute,("Model", "r_squared")]

        # Plots
        with plt.style.context("seaborn-white"):
            fig, axes = plt.subplots(figsize=(13,5), ncols=2)
            # ------------
            # Scatter plot
            # ------------
            # Left
            ax = axes[0]

            if show_attribute:
                xlabel = "$ŷ$({})".format(attribute)
                ylabel = "$y$({})".format(attribute)
            else:
                xlabel = "$ŷ$"
                ylabel = "$y$"

            plot_scatter(data=data, x="ŷ", y="y",  ax=ax, xlabel=xlabel, ylabel=ylabel, **_scatter_kws)#, cmap=cmap, vmin=vmin,vmax=vmax)
            ax.annotate(s="$R^2$ = %0.5f"%(r2), xy=(0, 1),  **_annot_kws)

            if show_reference_fit:
                y_actual = data["y"]
                slope, intercept, r_value, p_value, std_err = linregress(y_actual,y_actual)
                ylim = ax.get_ylim()
                x_grid = np.linspace(*ylim)
                ax.plot(x_grid, x_grid*slope + intercept, color="black", linestyle=":")
                ax.set_ylim(ylim)

            # Right
            ax = axes[1]
            plot_scatter(data=data, x="ŷ", y="residuals",  ax=ax, xlabel=xlabel, ylabel="Residuals", **_scatter_kws)#, cmap=cmap, vmin=vmin,vmax=vmax)
            ax.axhline(0, color="black", linestyle="-", linewidth=1, zorder=0)

            # Legend
            if legend is not None:
                ax.legend(*format_mpl_legend_handles(legend), **_legend_kws)
            # Title
            if title is not None:
                fig.suptitle(title, **_title_kws)

            return fig, ax

    # Save object to file
    def to_file(self, path:str, **kwargs):
        path = format_path(path)
        write_object(self, path, **kwargs)

# Mixed Linear Effects Models
class MixedEffectsLinearModel(CoreLinearModel):
    """
    import soothsayer as sy

    # Testing Mixed Linear Model
    X_iris, y_iris = sy.utils.get_iris_data(return_data=["X","y"], noise=47)
    # Training Data
    fixed_effects = ["petal_width"]
    X = X_iris.drop(fixed_effects, axis=1)
    Y = pd.concat([y_iris.to_frame(), X_iris[fixed_effects]], axis=1)
    # Model
    model = sy.regression.MixedEffectsLinearModel()
    # Fit
    model.fit(X=X, Y=Y, fixed_effects_variables=fixed_effects, random_effect_variable="Species")
    # Create association graph
    model.create_graph(tol_test=0.05)
    # Save object (Important b/c statsmodels 0.10.0 is broken) #https://github.com/statsmodels/statsmodels/issues/5899
    model.to_file("iris.model.pbz2")
    # Load it (This part wouldn't work w/ the above comment)
    model = sy.io.read_object("iris.model.pbz2")
    # Test to make sure it looks good
    model.plot_graph(show_node_labels=True)
    # Show the data
    model.synopsis_.head()

    """
    def __init__(
            self,
            name=None,
            attr_type=None,
            obsv_type=None,
            multiple_comparison_method="fdr",
            n_jobs=1,
            verbose=True,
            mixedlm_kws=dict(),
            ):
        super().__init__()
        self.name = name
        self.attr_type = attr_type
        self.obsv_type = obsv_type
        self.multiple_comparison_method = multiple_comparison_method
        self.n_jobs = n_jobs
        self.verbose = verbose
        self.fitted = False
        self.mixedlm_kws = dict()

    # Fit model for each attribute
    def fit(self, X:pd.DataFrame, Y:pd.DataFrame, fixed_effects_variables:list, random_effect_variable:str, formula:str=None, add_attributes_to_fixed_variables=False, categorical_variables="infer", fit_kws=dict(), assert_missing_values=True):
        # Should probably bring these methods into CoreLinearModel
        start_time = time.time()
        # Check inputs
        if assert_missing_values:
            assert np.all(X.notnull()), "X has missing values and it will throw downstream errors."
            assert np.all(Y.notnull()), "Y has missing values and it will throw downstream errors."
        assert len(set(X.columns) & set(Y.columns)) == 0, "There can't be an overlap of columns in X and Y."
        assert set(X.index) == set(Y.index), f"X.index and Y.index must have same values"
        Y = Y.loc[X.index,:]
        self.X_ = X.copy()
        self.Y_ = Y.copy()
        self.formula = formula
        self.residual_type = "resid"

        # Adding attributes to fixed variables
        self.add_attributes_to_fixed_variables = add_attributes_to_fixed_variables
        if self.add_attributes_to_fixed_variables:
            warnings.warn("Experimental: `add_attributes_to_fixed_variables`")

        # Categorical variables
        if categorical_variables == "infer":
            categorical_variables = list()
            for id_variable in self.Y_.columns:
                if self.Y_[id_variable].dtype == object:
                    categorical_variables.append(id_variable)
            if self.verbose:
                if bool(categorical_variables):
                    print("Inferred categorical variables from `Y`: {}".format(categorical_variables), file=sys.stderr)

        if bool(categorical_variables):
            assert is_nonstring_iterable(categorical_variables), "`categorical_variables` must be a non-string iterable if it is not None"
            for id_variable in categorical_variables:
                assert id_variable in self.Y_.columns, "Not all variables from `categorical_variables` are in `Y.columns`"
                self.Y_[id_variable] = self.Y_[id_variable].astype("category")

        # Encode the variables
        X_relabeled, encoding_X, decoding_X = self.relabel_attributes(X, attribute_prefix="attr_")
        Y_relabeled, encoding_Y, decoding_Y = self.relabel_attributes(Y, attribute_prefix="meta_")
        self.encoding_ = {**encoding_X, **encoding_Y}
        self.decoding_ = {**decoding_X, **decoding_Y}

        # Check variables
        self.attributes_ = X.index
        self.fixed_effect_variables_ = [*map(lambda x:self.encoding_[x], fixed_effects_variables)]
        self.random_effect_variable_ = self.encoding_[random_effect_variable]
        variables = fixed_effects_variables + [random_effect_variable]
        assert set(variables) <= set(Y.columns), f"{set(variables) - set(Y.columns)} not in `Y.columns`"

        # Run regression models
        args = {
            "X":X_relabeled,
            "Y":Y_relabeled,
            "formula":formula,
            "multiple_comparison_method":self.multiple_comparison_method,
            "residual_type":self.residual_type,
            "fit_kws":fit_kws,
        }
        if self.verbose:
            data = Parallel(n_jobs=self.n_jobs)(delayed(self._run_mixedlm)(query_attr, **args) for query_attr in tqdm(X_relabeled.columns, "Modeling each attribute"))
        else:
            data = Parallel(n_jobs=self.n_jobs)(delayed(self._run_mixedlm)(query_attr, **args) for query_attr in X_relabeled.columns)

        # Model results
        self.models_ = OrderedDict(zip(X.columns, map(lambda x:x[0], data)))
        self.model_results_ = OrderedDict(zip(X.columns, map(lambda x:x[1], data)))
        self.synopsis_ = pd.DataFrame([*map(lambda x:x[2], data)])
        self.synopsis_.index.name = "id_attribute"
        self.residuals_ = pd.DataFrame(OrderedDict(zip(X.columns, map(lambda x:x[3], data))))

        # Decode the variables
        self.fixed_effect_variables_ = [*map(lambda x:self.decoding_[x], self.fixed_effect_variables_)]
        self.random_effect_variable_ = self.decoding_[self.random_effect_variable_]

        # Duration
        self.fitted = True
        self.duration_ = format_duration(start_time)
        return self

    # Run statsmodels
    def _run_mixedlm(self, query_attr, X, Y, formula, multiple_comparison_method, residual_type, fit_kws):

        # Independent variables
        fixed_effect_variables = copy.deepcopy(self.fixed_effect_variables_)
        variables = fixed_effect_variables + [self.random_effect_variable_]

        index = sorted(set(X[query_attr].dropna().index) & set(Y[variables].dropna(how="any", axis=0).index))
        X = X.loc[index]
        Y = Y.loc[index]
        design_matrix = Y.loc[:, variables]
        if self.add_attributes_to_fixed_variables:
            idx_attrs_fixed = list(set(X.columns) - set([query_attr]))
            design_matrix = pd.concat([design_matrix, X.loc[:,idx_attrs_fixed]], axis=1)
            fixed_effect_variables += idx_attrs_fixed

        # Add dependent variable
        y_actual = pd.Series(X[query_attr].astype(float), name=self.decoding_[query_attr])
        design_matrix["y"] = y_actual

        # Formula
        if formula is None: # Not sure how this will do with custom formula and 'add_attributes_to_fixed_variables'
            formula = "y ~ " + " + ".join(fixed_effect_variables)

        # Model
        model = smf.mixedlm(formula=formula, data=design_matrix, groups=design_matrix[self.random_effect_variable_], **self.mixedlm_kws)
        model_results = model.fit(**fit_kws)
        setattr(model_results, "y_actual", y_actual)

        # Parse results
        synopsis = self.statsmodels_table_to_pandas(model_results, name=self.decoding_[query_attr], multiple_comparison_method=multiple_comparison_method, residual_type=residual_type)
        synopsis.index = synopsis.index.map(self.decode_index)

        return (model, model_results, synopsis, getattr(model_results, residual_type))

# Generalized Linear Models
class GeneralizedLinearModel(CoreLinearModel):
    def __init__(
            self,
            family=None,
            name=None,
            attr_type=None,
            obsv_type=None,
            multiple_comparison_method="fdr",
            n_jobs=1,
            verbose=True,
            glm_kws=dict(),
            ):
        super().__init__()
        self.name = name
        self.attr_type = attr_type
        self.obsv_type = obsv_type
        self.multiple_comparison_method = multiple_comparison_method
        self.n_jobs = n_jobs
        self.verbose = verbose
        self.fitted = False
        self.glm_kws = glm_kws
        if family is not None:
            self.glm_kws["family"] = family

    # Fit model for each attribute
    def fit(self, X:pd.DataFrame, Y:pd.DataFrame, fixed_effects_variables:list, formula:str=None, add_attributes_to_fixed_variables=False, categorical_variables="infer", residual_type="resid_response", fit_kws=dict(), assert_missing_values=True):
        start_time = time.time()
        # Check inputs
        if assert_missing_values:
            assert np.all(X.notnull()), "X has missing values and it will throw downstream errors."
            assert np.all(Y.notnull()), "Y has missing values and it will throw downstream errors."
        assert len(set(X.columns) & set(Y.columns)) == 0, "There can't be an overlap of columns in X and Y."
        assert set(X.index) == set(Y.index), f"X.index and Y.index must have same values"
        Y = Y.loc[X.index,:]
        self.X_ = X.copy()
        self.Y_ = Y.copy()
        self.formula = formula
        self.residual_type = residual_type

        # Adding attributes to fixed variables
        self.add_attributes_to_fixed_variables = add_attributes_to_fixed_variables
        if self.add_attributes_to_fixed_variables:
            warnings.warn("Experimental: `add_attributes_to_fixed_variables`")

        # Categorical variables
        if categorical_variables == "infer":
            categorical_variables = list()
            for id_variable in self.Y_.columns:
                if self.Y_[id_variable].dtype == object:
                    categorical_variables.append(id_variable)
            if self.verbose:
                if bool(categorical_variables):
                    print("Inferred categorical variables from `Y`: {}".format(categorical_variables), file=sys.stderr)

        if bool(categorical_variables):
            assert is_nonstring_iterable(categorical_variables), "`categorical_variables` must be a non-string iterable if it is not None"
            for id_variable in categorical_variables:
                assert id_variable in self.Y_.columns, "Not all variables from `categorical_variables` are in `Y.columns`"
                self.Y_[id_variable] = self.Y_[id_variable].astype("category")

        # Encode the variables
        X_relabeled, encoding_X, decoding_X = self.relabel_attributes(X, attribute_prefix="attr_")
        Y_relabeled, encoding_Y, decoding_Y = self.relabel_attributes(Y, attribute_prefix="meta_")
        self.encoding_ = {**encoding_X, **encoding_Y}
        self.decoding_ = {**decoding_X, **decoding_Y}

        # Check variables
        self.attributes_ = X.index
        self.fixed_effect_variables_ = [*map(lambda x:self.encoding_[x], fixed_effects_variables)]
        variables = fixed_effects_variables
        assert set(variables) <= set(Y.columns), f"{set(variables) - set(Y.columns)} not in `Y.columns`"

        # Run regression models
        args = {
            "X":X_relabeled,
            "Y":Y_relabeled,
            "formula":formula,
            "multiple_comparison_method":self.multiple_comparison_method,
            "residual_type":self.residual_type,
            "fit_kws":fit_kws,
        }
        if self.verbose:
            data = Parallel(n_jobs=self.n_jobs)(delayed(self._run_glm)(query_attr, **args) for query_attr in tqdm(X_relabeled.columns, "Modeling each attribute"))
        else:
            data = Parallel(n_jobs=self.n_jobs)(delayed(self._run_glm)(query_attr, **args) for query_attr in X_relabeled.columns)

        # Model results
        self.models_ = OrderedDict(zip(X.columns, map(lambda x:x[0], data)))
        self.model_results_ = OrderedDict(zip(X.columns, map(lambda x:x[1], data)))
        self.synopsis_ = pd.DataFrame([*map(lambda x:x[2], data)])
        self.synopsis_.index.name = "id_attribute"
        self.residuals_ = pd.DataFrame(OrderedDict(zip(X.columns, map(lambda x:x[3], data))))

        # Decode the variables
        self.fixed_effect_variables_ = [*map(lambda x:self.decoding_[x], self.fixed_effect_variables_)]

        # Duration
        self.fitted = True
        self.duration_ = format_duration(start_time)
        return self

    # Run statsmodels
    def _run_glm(self, query_attr, X, Y, formula, multiple_comparison_method, residual_type, fit_kws):

        # Independent variables
        fixed_effect_variables = copy.deepcopy(self.fixed_effect_variables_)
        variables = fixed_effect_variables

        # Handle missing values
        index = sorted(set(X[query_attr].dropna().index) & set(Y[variables].dropna(how="any", axis=0).index))
        X = X.loc[index]
        Y = Y.loc[index]

        design_matrix = Y.loc[:, variables]
        if self.add_attributes_to_fixed_variables:
            idx_attrs_fixed = list(set(X.columns) - set([query_attr]))
            design_matrix = pd.concat([design_matrix, X.loc[:,idx_attrs_fixed]], axis=1)
            fixed_effect_variables += idx_attrs_fixed


        # Add dependent variable
        y_actual = pd.Series(X[query_attr].astype(float), name=self.decoding_[query_attr])
        design_matrix["y"] = y_actual

        # Formula
        if formula is None: # Not sure how this will do with custom formula and 'add_attributes_to_fixed_variables'
            formula = "y ~ " + " + ".join(fixed_effect_variables)

        # Model
        model = smf.glm(formula=formula, data=design_matrix, **self.glm_kws)
        model_results = model.fit(**fit_kws)
        setattr(model_results, "y_actual", y_actual)
        # Parse results
        synopsis = self.statsmodels_table_to_pandas(model_results, name=self.decoding_[query_attr], multiple_comparison_method=multiple_comparison_method, residual_type=residual_type)
        synopsis.index = synopsis.index.map(self.decode_index)

        return (model, model_results, synopsis, getattr(model_results, residual_type))



# ==============================================================================
# Imports
# ==============================================================================
# Built-ins
import os, sys, time

import pandas as pd
import numpy as np

# Soothsayer
from ..r_wrappers import *
from soothsayer.utils import check_packages, is_query_class
# ==============================================================================
# R Imports
# ==============================================================================
if "rpy2" in sys.modules:
    from rpy2 import robjects as ro
    from rpy2 import rinterface as ri

    from rpy2.robjects.packages import importr
    try:
        from rpy2.rinterface import RRuntimeError
    except ImportError:
        from rpy2.rinterface_lib.embedded import RRuntimeError
    from rpy2.robjects import pandas2ri
    # pandas2ri.activate()
    R = ro.r
    NULL = ri.NULL
    #rinterface.set_writeconsole_regular(None)

# R packages
# dynamicTreeCut = R_package_retrieve("dynamicTreeCut")
# fastcluster = R_package_retrieve("fastcluster")
# ==============================================================================
# Exports
# ==============================================================================
__all__ = ["cutree_dynamic"]
# ==============================================================================
# dynamicTreeCut
# ==============================================================================
@check_packages(["dynamicTreeCut", "fastcluster"], language="r", import_into_backend=False)
def cutree_dynamic(kernel, cut_method="hybrid", method="ward", minClusterSize=20, name=None, **args):
    """
    dynamicTreeCut: cutreeDynamic
        function (dendro, cutHeight = NULL, minClusterSize = 20, method = "hybrid",
                   distM = NULL, de <...> it = deepSplit, minModuleSize = minClusterSize))
    """
    # Imports
    dynamicTreeCut = R_package_retrieve("dynamicTreeCut")
    fastcluster = R_package_retrieve("fastcluster")

    accepted_kernel_types = {"DataFrame", "Symmetric"}
    assert is_query_class(kernel, query=accepted_kernel_types), f"`kernel` type must be one of the following: {accepted_kernel_types}"
    # pd.DataFrame -> Symmetric
    if method == "ward":
        method = "ward.D2"
    if is_query_class(kernel, "Symmetric"):
        kernel = kernel.to_dense()
    if isinstance(kernel, pd.DataFrame):
        rkernel = pandas_to_rpy2(kernel)

    Z = fastcluster.hclust(R["as.dist"](rkernel), method=method)
    treecut_output = dynamicTreeCut.cutreeDynamic(dendro=Z, method=cut_method, distM=rkernel, minClusterSize = minClusterSize, **args)
    Se_treecut = pd.Series(rpy2_to_pandas(treecut_output), index=kernel.index, name=name).astype(int)
    return (Se_treecut - 1).sort_values() # Make outliers -1 and largest cluster 0
